from abc import ABC, abstractmethod
from collections import ChainMap
from copy import deepcopy
from datetime import datetime
from inspect import getmembers, isclass, isabstract, get_annotations
from pathlib import Path
import sys
from typing import Tuple, Any, Optional, Union, Callable, List
import warnings

from frozendict import frozendict
from nidaqmx.utils import flatten_channel_string
import numpy as np
from pydantic import BaseModel, Field, field_validator, ConfigDict, computed_field, model_validator
import tomli
import tomli_w

from .exceptions import (DualUserInterfaceError, HeadlessWarning, ImmutableInstanceWarning,
                         UnsupportedMicroscopeError, VersionBackwardCompatibilityError,
                         VersionForwardCompatibilityWarning, VersionBackwardCompatibilityWarning, UpdateVersionWarning)
from .meta import version
from .static import STATIC_PATH, SUPPORTED_MICROSCOPES, DEFAULT_GUI_SETTINGS
from .timing import Time, Milliseconds, Seconds
from .tools import (convert_permitted_types_to_required, FORMAT_TERMINAL, validate_extension, validate_filename, HAS,
                    EMPTY, random_shuffle, balanced_shuffle)


"""
The primary configuration is composed of a base template containing required information, and a tuple of
sub-configurations for components, adapters, and stimuli. The configuration template provides some methods for
verbose printing and user-oriented type-hinting. All configurations are based on the Pydantic BaseModel, which provides
a lot of functionality for free. Configurations can be written to or read from TOML.
"""


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CONFIG TEMPLATE & UTILITIES
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class ConfigTemplate(ABC, BaseModel):
    """
    Base Configuration Template. Extends pydantic base models with runtime type-checking when setting attributes and
    verbose type-hinting.
    """
    ###################################################################################################################
    # PYDANTIC CONFIGURATION..
    ###################################################################################################################
    model_config = ConfigDict(validate_assignment=True,
                              arbitrary_types_allowed=True,
                              strict=False,
                              extra="ignore",
                              frozen=False,
                              )

    ###################################################################################################################
    # CONFIGURATION CONSTANTS
    ###################################################################################################################
    version: str = Field(version, title="neurobeam version", description="Version of neurobeam package")

    def __str__(self) -> str:
        """
        Modified dunder method such that the printing is more verbose and easier for human consumption

        Prints the dataclass name and each of its parameters, their values, and associated types

        """
        string_to_print = ""
        for key, value in vars(self).items():
            if key == "components":
                if HAS(value):
                    for sub_config in value:
                        string_to_print += FORMAT_TERMINAL(f"{sub_config.__class__.__name__.replace('Config', '')}\n",
                                                           "component")
                        string_to_print += sub_config.__str__()
                else:
                    string_to_print += FORMAT_TERMINAL(f"{key}", "BLUE")
                    string_to_print += ":None\n"
            elif key == "adapters":
                if HAS(value):
                    for sub_config in value:
                        string_to_print += FORMAT_TERMINAL(f"{sub_config.__class__.__name__.replace('Config', '')}\n",
                                                           "adapter")
                        string_to_print += sub_config.__str__()
                else:
                    string_to_print += FORMAT_TERMINAL(f"{key}", "GREEN")
                    string_to_print += ":None\n"
            elif key in ("states", "transitions"):
                if HAS(value):
                    for sub_config in value:
                        string_to_print += FORMAT_TERMINAL(f"{sub_config.__class__.__name__.replace('Config', '')}\n",
                                                           "behavior")
                        string_to_print += sub_config.__str__()
                else:
                    string_to_print += FORMAT_TERMINAL(f"{key}", "ORANGE")
                    string_to_print += ":None\n"
            elif isinstance(value, (frozendict, dict)):
                # Not robust to nesting
                string_to_print += FORMAT_TERMINAL(f"{key}", "emphasis")
                string_to_print += ":"
                for nested_key in self.__dict__.get(key) if HAS(self.__dict__.get(key)) else {}:
                    string_to_print += f"\n\t{nested_key}: {self.__dict__.get(key).get(nested_key)}"
                string_to_print += "\n"
            else:
                string_to_print += FORMAT_TERMINAL(f"{key}", "emphasis")
                string_to_print += ":"
                string_to_print += f" {self.__dict__.get(key)}\n"

        return string_to_print

    @staticmethod
    @abstractmethod
    def __name__() -> str:
        return "Configuration Template"

    @staticmethod
    def _recursive_coerce(data: Any) -> dict:
        """
        Recursively coerce mutable to immutable

        :param data: The data to coerce
        :returns: Immutable data
        """
        if isinstance(data, (dict, frozendict)):
            data = frozendict({key: ConfigTemplate._recursive_coerce(value) for key, value in data.items()})
        elif isinstance(data, list):
            data = tuple([ConfigTemplate._recursive_coerce(value) for value in data])
        return data

    @classmethod
    def collect_annotations(cls: "ConfigTemplate") -> dict:
        """
        Collects annotations from all parent classes for type-checking

        :returns: Dictionary containing key-type pairs

        """
        return dict(ChainMap(*(get_annotations(cls_) for cls_ in cls.__mro__)))
    # It feels dumb to wrap with dict but for whatever reason it seems to be ordering it which is nice

    # noinspection PyNestedDecorators
    @field_validator("version")
    @classmethod
    def validate_version_compatibility(cls: "ConfigTemplate", v: str) -> str:
        """
        Validate the compatibility of the configuration version with the package version

        :param v: Configuration version
        :returns: Validated configuration version
        :raises VersionForwardCompatibilityWarning: Raised if the configuration major version is ahead of the package
            major version
        :raises VersionBackwardCompatibilityError: Raised if the configuration major version is behind the package
            major version
        :raises VersionBackwardCompatibilityWarning: Raised if the configuration patch version is behind the package
            patch version
        :raises UpdateVersionWarning: Raised if the configuration patch version is ahead of the package patch version
        """
        config_major, config_minor, config_patch = v.split(".")
        package_major, package_minor, package_patch = version.split(".")
        if int(config_major) < int(package_major):
            warnings.warn(VersionForwardCompatibilityWarning(v, version), stacklevel=2)
        elif int(config_major) > int(package_major):
            raise VersionBackwardCompatibilityError(v, version)
        elif int(config_minor) > int(package_minor):
            warnings.warn(VersionBackwardCompatibilityWarning(v, version), stacklevel=2)
        elif int(config_patch) > int(package_patch):
            warnings.warn(UpdateVersionWarning(v, version), stacklevel=2)
        return v

    @classmethod
    def __toml_decode__(cls, data: dict) -> "ConfigTemplate":
        """
        Decode a TOML dictionary into a configuration

        :param data: The dictionary containing the serialized configuration
        :returns: A deserialized and instantiated configuration template

        .. seealso:: :func:`decode_toml <neurobeam.configs.decode_toml>`
        """
        return cls(**decode_toml(data))

    def hint_types(self) -> None:
        """
        Print type hints for the configuration
        """
        type_hints = ""
        annotations_ = self.collect_annotations()
        for key_value, key_type in zip(vars(self).items(), annotations_.items()):
            key, value = key_value
            key_, type_ = key_type
            assert key == key_
            if key == "components":
                if HAS(value):
                    for sub_config in value:
                        type_hints += FORMAT_TERMINAL(f"{sub_config.__class__.__name__.replace('Config', '')}\n",
                                                      "component")
                        type_hints += sub_config.hint_types()
            elif key == "adapters":
                if HAS(value):
                    for sub_config in value:
                        type_hints += FORMAT_TERMINAL(f"{sub_config.__class__.__name__.replace('Config', '')}\n",
                                                      "adapter")
                        type_hints += sub_config.hint_types()
            elif key in ("states", "transitions"):
                if HAS(value):
                    for sub_config in value:
                        type_hints += FORMAT_TERMINAL(f"{sub_config.__class__.__name__.replace('Config', '')}\n",
                                                      "behavior")
                        type_hints += sub_config.hint_types()
            elif "typing.Optional" in f"{type_}":
                type_hints += FORMAT_TERMINAL(f" Optional[{type_.__args__[0].__name__}]\n", "type")
            else:
                type_hints += FORMAT_TERMINAL(f"{key}", "emphasis")
                type_hints += ":"
                type_hints += FORMAT_TERMINAL(f" {type_.__name__}\n", "type")
        print(type_hints)

    def save(self, save_location: Union[str, Path]) -> None:
        """
        Save the configuration to a TOML file
        """
        save_config(save_location, self)

    @model_validator(mode="before")
    def _coerce_mutable(self) -> "ConfigTemplate":
        """
        Coerce mutable fields to immutable types

        :returns: Immutable configuration
        """
        for key, value in self.items():
            if isinstance(value, (dict, frozendict, list)):
                self[key] = ConfigTemplate._recursive_coerce(value)
        return self

    def __toml_encode__(self) -> dict:
        """
        Recursively encode the configuration into a TOML-serializable dictionary

        :returns: The dictionary containing the serialized configuration
        """
        return encode_toml(vars(self))


class _ConfigTemplateFactory:
    """
    Config Template Factory
    """
    #: dict: registry of component template models
    __registry = {}

    @classmethod
    def register_template(cls):  # noqa: ANN001, ANN201, ANN206
        """
        Register a config template with the factory.
        """
        def register_model(config_template):  # noqa: ANN001, ANN201, ANN206
            factory_key = config_template.__fields__.get("factory_key").default
            cls.__registry[factory_key] = config_template
            return config_template
        return register_model

    @classmethod
    def is_config_template(cls, data: Any) -> bool:
        """
        Check if a config template is in the registry
        """
        if isinstance(data, dict) and "factory_key" in data:
            return data.get("factory_key") in cls.__registry

    @classmethod
    def construct_model(cls, data: dict) -> "ConfigTemplate":
        """
        Construct a config template from the deserialized dictionary and the factory key
        """
        factory_key = data.get("factory_key")
        if factory_key not in cls.__registry:
            raise KeyError(f"Factory key {factory_key} not found in registry")
        return cls.__registry[factory_key](**data)

    @classmethod
    def get_config_template(cls, factory_key: str) -> "ConfigTemplate":
        """
        Get a config template from the registry
        """
        if factory_key not in cls.__registry:
            raise KeyError(f"Factory key {factory_key} not found in registry")
        return cls.__registry[factory_key]


class HoistedConfig:
    """
    :class:`Config <neurobeam.configs.Config>` with a specific sub-configuration hoisted to the top-level of the namespace

    :param config: The top-level namespace
    :param key: The field to hoist
    :param idx: The index within the field to hoist

    .. seealso:: :class:`Config <neurobeam.configs.Config>`
    .. seealso:: :class:`ConfigTemplate <neurobeam.configs.ConfigTemplate>`
    """
    def __init__(self, config: "Config", key: str, idx: Optional[int] = None):
        #: Config: The top-level namespace, a Config instance
        self.config = config
        #: ConfigTemplate: The hoisted sub-configuration
        self.hoisted = getattr(config, key)[idx] if HAS(idx) else getattr(config, key)

    def __getattr__(self, item: str) -> Any:
        """
        Get an attribute from the hoisted sub-configuration, and if it doesn't exist, get it from the top-level

        :param item: The attribute to get
        :returns: The attribute from the hoisted sub-configuration
        :raises AttributeError: Raised if the attribute is not found in the hoisted or top-level configuration
        """
        try:
            return getattr(self.hoisted, item)
        except AttributeError:
            return getattr(self.config, item)


def decode_toml(data: dict) -> Union[dict, "ConfigTemplate"]:
    """
    Recursively decode a TOML-serialized dictionary into a configuration

    :param data: The dictionary to decode
    :returns: The deserialized dictionary
    """
    decoding = deepcopy(data)  # Avoid modifying the original dictionary
    for key, value in decoding.items():
        # Check if component template
        if _ConfigTemplateFactory.is_config_template(value):
            return _ConfigTemplateFactory.construct_model(decode_toml(value))
        # Check if sub-configuration
        elif key in ("components", "adapters", "states", "transitions", "stimuli") and isinstance(value, dict):
            decoding[key] = tuple([decode_toml({key: sub_value}) for key, sub_value in value.items()])
        # check if frozendict
        elif isinstance(value, dict):
            decoding[key] = frozendict(value)
        elif isinstance(value, list):
            if len(value) == 2 and isinstance(value[0], (int, float)) and isinstance(value[1], str):
                decoding[key] = Time.__toml_decode__(value)
            else:
                decoding[key] = tuple(value)
        # check if None
        elif value == "NULL":
            decoding[key] = None
        else:
            pass
    return decoding


def encode_toml(data: dict) -> dict:
    """
    Recursively encode a dictionary into a TOML-serializable dictionary

    :param data: The dictionary to encode
    :returns: The TOML-serializable dictionary
    """
    encoding = deepcopy(data)  # Avoid modifying the original dictionary
    for key, value in encoding.items():
        if isinstance(value, ConfigTemplate):
            encoding[key] = encode_toml(vars(value))
        elif isinstance(value, Time):
            encoding[key] = value.__toml_encode__()
        # for config template tuples/sets
        elif isinstance(value, (tuple, set)) and all(isinstance(sub_value, ConfigTemplate) for sub_value in value):
            encoding[key] = {f"{idx}": encode_toml(vars(sub_value)) for idx, sub_value in enumerate(value)}
        # for other tuples/sets
        elif isinstance(value, (tuple, set)) and not all(isinstance(sub_value, ConfigTemplate) for sub_value in value):
            encoding[key] = [str(val) for val in value]
        elif isinstance(value, np.ndarray):
            encoding[key] = [str(val) for val in value.tolist()]
        elif isinstance(value, frozendict):
            encoding[key] = dict(value)
        elif EMPTY(value):
            encoding[key] = "NULL"
        else:
            encoding[key] = str(value)
    return encoding


@convert_permitted_types_to_required(permitted=(str, Path), required=Path, pos=0, key="path")
@validate_extension(required_extension=".toml", pos=0, key="path")
def load_config(path: Path) -> "ConfigTemplate":
    """
    Load a configuration from a TOML file

    :param path: The path to load configuration from
    :returns: The configuration loaded from the file
    """
    with open(path, "rb") as file:
        data = tomli.load(file)
        if "factory_key" in data:
            cls = _ConfigTemplateFactory.get_config_template(data.get("factory_key"))
            return cls.__toml_decode__(data)
        else:
            return Config.__toml_decode__(data)


@convert_permitted_types_to_required(permitted=(str, Path), required=Path, pos=0, key="path")
@validate_filename(pos=0, key="path")
@validate_extension(required_extension=".toml", pos=0, key="path")
def save_config(path: Path, config: "ConfigTemplate") -> None:
    """
    Saves a configuration to a TOML file

    :param path: The path to save configuration
    :param config: The configuration to save
    """

    with open(path, "wb") as file:
        tomli_w.dump(config.__toml_encode__(), file)


@convert_permitted_types_to_required(permitted=(str, Path), required=Path, pos=0, key="path")
def generate_default_configs(path: Path = STATIC_PATH, target: Optional[Union[str, "ConfigTemplate"]] = None) -> None:
    """
    Generate default configurations

    :param path: The path to save configurations
    :param target: The name of a specific configuration to generate default for (e.g., "Config")
    """
    config_module = sys.modules[__name__]
    if isclass(target) and issubclass(target, ConfigTemplate) and not isabstract(target):
        # noinspection PyCallingNonCallable
        save_config(path.joinpath(f"{target.__name__}.toml"), target())
    else:
        for name, cls in getmembers(config_module, isclass):
            if issubclass(cls, ConfigTemplate) and not isabstract(cls):
                if target is None or target == name:
                    save_config(path.joinpath(f"{name}.toml"), cls())


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CALLBACK REQUESTS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@_ConfigTemplateFactory.register_template()
class CallbackRequestConfig(ConfigTemplate):

    factory_key: str = Field("CallbackRequest", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    key: Union[str, Tuple[str, ...]] = Field("callback",
                                             title="key",
                                             description="Key/s for the callback request")

    approximate: bool = Field(False,
                              title="approximate",
                              description="Boolean indicating whether to retrieve the callback as an approximate match")

    required: bool = Field(False, title="required",
                           description="Boolean indicating whether the callback is required")

    arguments: Optional[Union[frozendict[str, Any], Tuple[frozendict[str, Any]]]] =\
        Field(None, title="arguments", description="Any keyword arguments for the callback")

    wildcard: Optional[Union[Tuple[str, Any], Tuple[Tuple[str, Any], ...]]] \
        = Field(None, title="wildcard", description="wildcard")

    @staticmethod
    def __name__() -> str:
        return "Callback Request Configuration"


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// COMPONENT SUB-CONFIGURATIONS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


# noinspection PyNestedDecorators
@_ConfigTemplateFactory.register_template()
class CameraConfig(ConfigTemplate):
    """
    Sub-configuration for creating camera components.

    .. seealso:: :class:`ConfigTemplate <neurobeam.configs.ConfigTemplate>`
    .. seealso:: :class:`Camera <neurobeam.hardware.video.Camera>`
    .. seealso:: :class:`VideoFeed <prove.hardware.video.VideoFeed>`
    """
    factory_key: str = Field("Camera", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    # note that there isn't a clear way to get the opencv id using the MSMF backend & it's a bit of a pain to set up,
    # however microsoft recommends using the MSMF backend and it supports hardware-acceleration
    camera_idx: int = Field(0, title="camera_idx", description="OpenCV index of camera to use", ge=0)

    save_video: bool = Field(True,
                             title="save_feed",
                             description="Boolean indicating whether to save camera feed to file")

    height: int = Field(480, title="height", description="Height of camera frame", ge=1)

    width: int = Field(640, title="width", description="Width of camera frame", ge=1)

    fps: float = Field(30.0, title="fps", description="Frames per second of camera", gt=0)

    @staticmethod
    def __name__() -> str:
        return "Camera Configuration"


# noinspection PyNestedDecorators
@_ConfigTemplateFactory.register_template()
class DAQConfig(ConfigTemplate):
    """
    Sub-configuration for creating DAQ components.

    .. seealso:: :class:`ConfigTemplate <neurobeam.configs.ConfigTemplate>`
    .. seealso:: :class:`DAQ <neurobeam.hardware.daq.DAQ>`
    """
    factory_key: str = Field("DAQ", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    device_id: str = Field("SimDAQ", title="device_id", description="Name of National Instruments DAQ")

    timeout: float = Field(1.0,
                           title="timeout",
                           description="Timeout for all DAQ interactions (s)",
                           gt=0.0)

    sampling_frequency: int = Field(1000,
                                    title="sampling_frequency",
                                    description="Sampling frequency (Hz) for hardware-timed DAQ streams",
                                    ge=1,
                                    le=10000)

    buffer_frequency: int = Field(100,
                                  title="buffer_frequency",
                                  description="Buffering frequency (Hz) for hardware-timed DAQ streams")

    digital_analog_sampling_ratio: float = (
        Field(0.5,
              title="digital_analog_sampling_ratio",
              description="Ratio of software-timed digital to hardware-timed analog samples",
              gt=0.0,
              lt=1.0))

    analog_input_index: frozendict[str, int] = (
        Field(frozendict({"imaging_start": 0,
                          "linear_actuator_position": 1,
                          "wheel_position": 2,
                          "burrow_tension": 3,
                          "locomotion": 4}),
              title="analog_input_index",
              description="Mapping of analog input keys to physical indices"))

    analog_input_range: Tuple[float, float] = Field(default=(-10.0, 10.0),
                                                    title="analog_input_range",
                                                    description="Range of permissible analog input voltage ranges")

    analog_output_index: frozendict[str, int] = (
        Field(frozendict({"stimulus": 0, "linear_actuator": 1}),
              title="analog_output_index",
              description="Mapping of analog output keys to physical indices"))

    analog_output_range: frozendict[str, Tuple[float, float]] = (
        Field(frozendict({"stimulus": (0.0, 5.0), "linear_actuator": (0.0, 5.0)}),
              title="analog_output_range",
              description="Mapping of analog output keys to permissible output voltage ranges"))

    digital_input_index: frozendict[str, Tuple[int, int]] = (
        Field(frozendict(
            {"burrow_sensor": (0, 0),
             "reward_sucrose": (0, 1),
             "reward_water": (0, 2),
             "licking_sucrose": (0, 3),
             "licking_water": (0, 4)}),
            title="digital_input_index",
            description="Mapping of digital input keys to physical indices"))

    digital_output_index: frozendict[str, Tuple[int, int]] = (
        Field(frozendict({
            "burrow_sensor_driver": (0, 5),
            "event_indicator": (0, 6),
            "trial_indicator": (0, 7),
            "permit_water_reward": (1, 0),
            "permit_sucrose_reward": (1, 1),
            "rotate_spouts": (1, 2),
            "remove_spouts": (1, 3),
            "wet_start": (2, 0)
        }),
            title="digital_output_index",
            description="Mapping of digital output keys to physical indices"))

    @computed_field
    @property
    def analog_buffer_size(self) -> int:
        """
        Size of analog buffer in samples

        """

        return self.sampling_frequency // self.buffer_frequency

    @computed_field
    @property
    def digital_buffer_size(self) -> int:
        """
        Size of digital buffer in samples

        """
        return round(self.sampling_frequency / self.buffer_frequency)

    @computed_field
    @property
    def analog_output_channels(self) -> dict:
        return {key: "".join([self.device_id, "/ao", str(self.analog_output_index.get(key))]) for key in
                self.analog_output_index.keys()}

    @computed_field
    @property
    def analog_output_cids(self) -> Tuple[str, ...]:
        return self.analog_output_index.keys()

    @computed_field
    @property
    def num_analog_output_channels(self) -> int:
        return len(self.analog_output_index.keys())

    @computed_field
    @property
    def flattened_analog_outputs(self) -> str:
        return flatten_channel_string([self.analog_output_channels.get(key)
                                       for key in self.analog_output_channels.keys()])

    @computed_field
    @property
    def analog_input_channels(self) -> dict:
        return {key: "".join([self.device_id, "/ai", str(self.analog_input_index.get(key))]) for key in
                self.analog_input_index.keys()}

    @computed_field
    @property
    def analog_input_cids(self) -> Tuple[str, ...]:
        return self.analog_input_index.keys()

    @computed_field
    @property
    def num_analog_input_channels(self) -> int:
        return len(self.analog_input_index.keys())

    @computed_field
    @property
    def flattened_analog_inputs(self) -> str:
        return flatten_channel_string([self.analog_input_channels.get(key)
                                       for key in self.analog_input_channels.keys()])

    @computed_field
    @property
    def digital_input_channels(self) -> dict:
        return {key: "".join([
            self.device_id, "/port", str(self.digital_input_index.get(key)[0]), "/line",
            str(self.digital_input_index.get(key)[1])]) for key in self.digital_input_index.keys()}

    @computed_field
    @property
    def digital_input_cids(self) -> Tuple[str, ...]:
        return self.digital_input_index.keys()

    @computed_field
    @property
    def num_digital_input_channels(self) -> int:
        return len(self.digital_input_index.keys())

    @computed_field
    @property
    def flattened_digital_inputs(self) -> str:
        return flatten_channel_string([self.digital_input_channels.get(
            key) for key in self.digital_input_channels.keys()])

    @computed_field
    @property
    def digital_output_channels(self) -> dict:
        return {key: "".join([self.device_id, "/port", str(self.digital_output_index.get(key)[0]), "/line",
                              str(self.digital_output_index.get(key)[1])]) for key in self.digital_output_index.keys()}

    @computed_field
    @property
    def num_digital_output_channels(self) -> int:
        return len(self.digital_output_index.keys())

    @computed_field
    @property
    def flattened_digital_outputs(self) -> str:
        return flatten_channel_string([self.digital_output_channels.get(key)
                                       for key in self.digital_output_channels.keys()])

    @staticmethod
    def __name__() -> str:
        return "Hardware Configuration"

    @model_validator(mode="after")
    def _validate_analog_output_mapping(self) -> "DAQConfig":
        assert set(self.analog_output_index.keys()) == set(self.analog_output_range.keys()), \
            "Analog output index and range mappings must have the same keys"
        return self

    @model_validator(mode="after")
    def _validate_index_mapping(self) -> "DAQConfig":
        mappings = [self.analog_input_index,
                    self.analog_output_index,
                    self.digital_input_index,
                    self.digital_output_index]
        keys = [key for mapping in mappings for key in mapping.keys()]
        assert len(keys) == len(set(keys)), "Keys must be unique"
        return self

    @model_validator(mode="after")
    def _validate_effective_sampling(self) -> "DAQConfig":
        if self.sampling_frequency % self.buffer_frequency != 0:
            raise AssertionError(f"Buffer size must be int: {self.sampling_frequency=}, {self.buffer_frequency=}, "
                                 f"buffer size={self.sampling_frequency / self.buffer_frequency}")
        if self.buffer_frequency > self.sampling_frequency:
            raise AssertionError(f"Buffer frequency must be less than or equal to sampling frequency: "
                                 f"{self.sampling_frequency=}, {self.buffer_frequency=}")

        return self


# noinspection PyNestedDecorators
@_ConfigTemplateFactory.register_template()
class MicroscopeConfig(ConfigTemplate):
    """
    Sub-configuration for creating microscope components.

    .. seealso:: :class:`ConfigTemplate <neurobeam.configs.ConfigTemplate>`
    .. seealso:: :class:`Microscope <neurobeam.hardware.microscopes.Microscope>`
    """
    factory_key: str = Field("Microscope", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    software: str = Field("PrairieView",
                          title="brand",
                          description="Name of the microscope software")

    realtime_mode: bool = Field(False,
                                title="realtime_mode",
                                description="Boolean indicating whether to interface in realtime mode")

    motion_correction: bool = Field(False,
                                    title="motion_correction",
                                    description="Boolean indicating whether to motion correct images online")

    identify_rois: bool = Field(False,
                                title="identify_rois",
                                description="Boolean indicating whether to identify ROIs online")

    extract_fluorescence: bool = Field(False,
                                       title="extract_fluorescence",
                                       description="Boolean indicating whether to extract fluorescence online")

    infer_spikes: bool = Field(False,
                               title="infer_spikes",
                               description="Boolean indicating whether to infer spikes online")

    monitor_ensembles: bool = Field(False,
                                    title="monitor_ensembles",
                                    description="Boolean indicating whether to monitor ensemble activity online")

    @staticmethod
    def __name__() -> str:
        return "Microscope Configuration"

    @field_validator("software")
    @classmethod
    def _validate_brand(cls: "MicroscopeConfig", v: str) -> str:
        """
        Validate the microscope software brand

        :param v: The microscope software brand
        :returns: Validated microscope software brand
        """
        if not any((v.lower() in scope.lower() for scope in SUPPORTED_MICROSCOPES)):
            raise UnsupportedMicroscopeError(v)
        return v

    @model_validator(mode="after")
    def _warn_implemented_settings(self) -> "MicroscopeConfig":
        """
        Warn if the selected settings are not implemented in this version of the software

        :returns: Validated settings
        """
        for key, value in vars(self).items():
            if key in ("motion_correction",
                       "identify_rois",
                       "extract_fluorescence",
                       "infer_spikes",
                       "monitor_ensembles",
                       "realtime_mode"):
                if value:
                    warnings.warn(VersionBackwardCompatibilityWarning(key, value), stacklevel=2)
        return self


# noinspection PyNestedDecorators
@_ConfigTemplateFactory.register_template()
class SLMConfig(ConfigTemplate):
    """
    Sub-configuration for creating photostimulation components.

    .. seealso:: :class:`ConfigTemplate <neurobeam.configs.ConfigTemplate>`
    .. seealso:: :class:`SLM <neurobeam.hardware.slm.SLM>`
    """
    factory_key: str = Field("SLM", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    brand: str = Field("Meadowlark", title="brand", description="Name of the spatial light modulator brand")

    realtime_mode: bool = Field(False,
                                title="realtime_mode",
                                description="Boolean indicating whether to interface in realtime mode")

    @staticmethod
    def __name__() -> str:
        return "SLM Configuration"

    @model_validator(mode="after")
    def _warn_implemented_settings(self) -> "SLMConfig":
        """
        Warn if the selected settings are not implemented in this version of the software

        :returns: Validated settings
        """
        for key, value in vars(self).items():
            if key in ("realtime_mode", ):
                if value:
                    warnings.warn(VersionBackwardCompatibilityWarning(key, value), stacklevel=2)
        return self


# noinspection PyNestedDecorators
@_ConfigTemplateFactory.register_template()
class SpeakerConfig(ConfigTemplate):
    """
    Sub-configuration for creating speaker components.

    .. seealso:: :class:`ConfigTemplate <neurobeam.configs.ConfigTemplate>`
    .. seealso:: :class:`Speaker <neurobeam.hardware.speakers.Speaker>`
    """
    factory_key: str = Field("Speaker", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    speaker_id: str = Field("Speakers (High Definition", title="speaker_id",
                            description="Name of audio output device")

    volume: float = Field(1.0, title="volume", description="Volume of audio output", ge=0.0, le=1.0)

    lut: Optional[frozendict[str, float]] = Field(None, title="lut",
                                                  description="Lookup table for audio intensity")

    @staticmethod
    def __name__() -> str:
        return "Speaker Configuration"

    @field_validator("lut")
    @classmethod
    def _validate_lut(cls: "SpeakerConfig", v: Union[Path, frozendict[str, np.ndarray]]) \
            -> Union[Path, frozendict[str, np.ndarray]]:
        """
        Validate the lookup table for audio intensity

        :param v: The lookup table
        :returns: The validated lookup table
        """
        if isinstance(v, Path):
            v = np.load(v, allow_pickle=True).item()
        if HAS(v):
            return {str(key): value for key, value in v.items()}


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// ADAPTER SUB-CONFIGURATIONS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@_ConfigTemplateFactory.register_template()
class AnalogPulseConfig(ConfigTemplate):
    factory_key: str = Field("AnalogPulse", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    name: str = Field("Analog Pulse", title="name", description="Name of the analog pulse")

    device_id: str = Field("SimDAQ",
                           title="device_id",
                           description="Name of DAQ implementing the linear actuator")

    analog_pulse_channel_key: str = Field("analog_pulse",
                                          title="analog_pulse_channel_key",
                                          description="Key for the pulse output channel")

    @staticmethod
    def __name__() -> str:
        return "Analog Pulse Configuration"


@_ConfigTemplateFactory.register_template()
class AudioStimulusConfig(ConfigTemplate):
    """
    Sub-configuration for creating auditory stimuli adapters
    """
    factory_key: str = Field("AudioStimulus", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    name: str = Field("Audio Stimulus", title="name", description="Name of the audio stimulus")

    device_id: str = Field("Digital",
                           title="device_id",
                           description="Name of Speaker implementing the audio stimulus")

    style: str = Field("white_noise", title="style", description="Type of audio stimulus")

    duration: Time = Field(Seconds(1), title="duration", description="Duration of the audio stimulus")

    options: Optional[frozendict[str, Any]] = Field(None,
                                                    title="options",
                                                    description="Any additional keyword options for the stimulus")

    @staticmethod
    def __name__() -> str:
        return "Audio Stimulus Configuration"


@_ConfigTemplateFactory.register_template()
class LinearActuatorConfig(ConfigTemplate):
    """
    Sub-configuration for creating linear actuator adapters.
    """
    factory_key: str = Field("LinearActuator", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    name: str = Field("Linear Actuator", title="name", description="Name of the linear actuator")

    device_id: str = Field("SimDAQ",
                           title="device_id",
                           description="Name of DAQ implementing the linear actuator")

    command_channel_key: str = Field("linear_actuator",
                                     title="command_channel_key",
                                     description="Key for the linear actuator output channel")

    physical_range: Tuple[float, float] = Field((0.0, 50.0),
                                                title="physical_range",
                                                description="Physical range of the linear actuator")

    physical_units: str = Field("mm",
                                title="physical_units",
                                description="Physical units of the linear actuator range")

    @staticmethod
    def __name__() -> str:
        return "Linear Actuator Configuration"


@_ConfigTemplateFactory.register_template()
class StimulusTriggerConfig(ConfigTemplate):
    """
    Sub-configuration for creating stimulus trigger adapters.
    """
    factory_key: str = Field("StimulusTrigger", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    name: str = Field("Stimulus Trigger", title="name", description="Name of the stimulus trigger")

    stimulus_name: str = Field("AudioStimulus",
                               title="stimulus_name",
                               description="Name of the stimulus to trigger")

    relative_onset: Time = Field(Seconds(0), title="relative_onset", description="Relative onset of the trigger")

    @staticmethod
    def __name__() -> str:
        return "Stimulus Trigger Configuration"


@_ConfigTemplateFactory.register_template()
class DigitalTriggerConfig(ConfigTemplate):
    """
    Sub-configuration for creating digital trigger adapters.
    """
    factory_key: str = Field("DigitalTrigger", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor. "
                                         "Identical to the name of the implementing class")

    name: str = Field("Digital Trigger", title="name", description="Name of the digital trigger")

    device_id: str = Field("SimDAQ",
                           title="device_id",
                           description="Name of DAQ implementing the digital trigger")

    digital_trigger_channel_key: str = Field("digital_trigger",
                                             title="digital_trigger_channel_key",
                                             description="Key for the digital trigger channel")

    @staticmethod
    def __name__() -> str:
        return "Digital Trigger Configuration"


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// BEHAVIOR CONFIGURATION & SUB-CONFIGURATIONS (TECHNICALLY A COMPONENT SUB-CONFIGURATION WITH SUB-CONFIGURATIONS)
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@_ConfigTemplateFactory.register_template()
class StateConfig(ConfigTemplate):
    factory_key: str = Field("State", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor."
                                         "Identical to the name of the implementing class")

    name: str = Field("State", title="name", description="Name of the state")

    duration: Optional[Time] = Field(None, title="duration", description="Duration of the state")

    entry_callbacks: Optional[Union[CallbackRequestConfig, Callable, Tuple[Callable, ...]]] \
        = Field(None, title="entry_callbacks", description="Entry callbacks")
    exit_callbacks: Optional[Union[CallbackRequestConfig, Callable, Tuple[Callable, ...]]] \
        = Field(None, title="exit_callbacks", description="Exit callbacks")

    triggers: Optional[List[Tuple[str, Tuple[str, str]]]] = Field(None,
                                                                  title="triggers",
                                                                  description="Triggers for the state")

    special: Optional[str] = Field(None, title="special", description="keyword of a special state implementation")

    options: Optional[frozendict[str, Any]] = Field(None,
                                                    title="options",
                                                    description="Additional options for a special state")

    @staticmethod
    def __name__() -> str:
        return "State Configuration"


@_ConfigTemplateFactory.register_template()
class TransitionConfig(ConfigTemplate):
    factory_key: str = Field("Transition", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor."
                                         "Identical to the name of the implementing class")

    name: str = Field("Auto", title="name", description="Name of the transition")

    origin: str = Field("Setup", title="origin", description="Name of the origin state")

    targets: Union[str, Tuple[str, ...]] = Field("End",
                                                 title="targets",
                                                 description="Names of the target state/s")

    order: Optional[Tuple[int, ...]] = Field(None, title="order", description="Order of the target states")

    final: Optional[str] = Field(None, title="final", description="Name of the final state")

    shuffle_method: Optional[str] = Field(None,
                                          title="shuffle_order",
                                          description="Boolean indicating whether to shuffle the order")
    _shuffled = False

    @staticmethod
    def __name__() -> str:
        return "Transition Configuration"

    @property
    def shuffled(self):
        return self._shuffled

    @shuffled.setter
    def shuffled(self, value: bool):
        self._shuffled = value

    @model_validator(mode="after")
    def _shuffle_order(self) -> "TransitionConfig":
        if HAS(self.shuffle_method) and not self.shuffled:
            if self.shuffle_method == "random":
                self._shuffled = True
                self.order = tuple(random_shuffle(self.order))
            elif self.shuffle_method == "balanced":
                self._shuffled = True
                self.order = tuple(balanced_shuffle(self.order))
            elif self.shuffle_method == "resampled":
                raise NotImplementedError
            else:
                raise NotImplementedError

        return self


@_ConfigTemplateFactory.register_template()
class BehaviorConfig(ConfigTemplate):
    factory_key: str = Field("Behavior", title="factory_key", allow_mutation=False,
                             description="Key used to locate constructor."
                                         "Identical to the name of the implementing class")

    name: str = Field("Behavioral Task", title="name", description="Name of the behavioral task")

    states: Optional[Tuple[StateConfig, ...]] = Field(None, title="states", description="State configurations")

    transitions: Optional[Tuple[TransitionConfig, ...]] = Field(None, title="transitions",
                                                                description="Transition configurations")

    @staticmethod
    def __name__() -> str:
        return "Behavior Configuration"

    def add_state(self, state: "StateConfig") -> None:
        """
        Add a state configuration to the behavior configuration

        :param state: Configuration of the state to add to the behavior

        .. seealso:: :class:`StateConfig <neurobeam.configs.StateConfig>`
        """
        if self.states is None:
            self.states = (state, )
        else:
            self.states += (state, )

    def add_transition(self, transition: "TransitionConfig") -> None:
        """
        Add a transition configuration to the behavior configuration

        :param transition: Configuration of the transition to add to the behavior

        .. seealso:: :class:`TransitionConfig <neurobeam.configs.TransitionConfig>`
        """
        if self.transitions is None:
            self.transitions = (transition, )
        else:
            self.transitions += (transition, )


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROBE CONFIGURATION
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


# noinspection PyNestedDecorators
class Config(ConfigTemplate):
    """
    Primary configuration for the neurobeam package. Contains experimental settings, metadata, and (potentially)
    component and adapter sub-configurations.
    """
    ###################################################################################################################
    # EXPERIMENT METADATA (MUTABLE)
    ###################################################################################################################
    subject: str = Field("subject",
                         title="subject",
                         description="Name or ID of the experimental subject")

    experimenter: str = Field("John Doe",
                              title="experimenter",
                              description="Name of the scientist conducting experiment")

    session: int = Field(1,
                         title="session",
                         description="Experiment session number (1-indexed)",
                         ge=0)

    study: str = Field("study", title="study", description="Name or ID of the study")

    condition: str = Field("condition",
                           title="condition",
                           description="Name or ID of the experimental condition")

    notes: Optional[frozendict[str, Any]] = Field(None,
                                                  title="notes",
                                                  description="Notes about the experiment")

    ###################################################################################################################
    # EXPERIMENT METADATA (CONSTANT)
    ###################################################################################################################
    # noinspection PyTypeHints
    date: str = Field(datetime.now().strftime("%Y-%m-%d"),
                      title="date",
                      description="Date of experiment")

    # noinspection PyTypeHints
    timestamp: str = Field(datetime.now().strftime("%H:%M:%S"),
                           title="timestamp",
                           description="Time of experiment")

    ###################################################################################################################
    # EXPERIMENT CONFIGURATION
    ###################################################################################################################
    save_directory: Path = Field(Path.cwd(),
                                 title="save_directory",
                                 description="Directory to save experiment data")

    use_console: bool = Field(True,
                              title="use_console",
                              description="Boolean indicating whether to print progress to console")

    use_gui: bool = Field(False,
                          title="use_gui",
                          description="Boolean indicating whether to launch graphical user interface for experiment")

    gui_settings: Path = Field(DEFAULT_GUI_SETTINGS, title="gui_settings", description="Path to GUI settings file")

    reporting_frequency: float = Field(1.0,
                                       title="reporting_frequency",
                                       description="Frequency of progress reporting")

    ###################################################################################################################
    # SUB-CONFIGURATIONS
    ###################################################################################################################
    components: Optional[Tuple["ConfigTemplate", ...]] = Field(None,
                                                               title="components",
                                                               description="Sub-configurations for neurobeam components")

    adapters: Optional[Tuple["ConfigTemplate", ...]] = Field(None,
                                                             title="adapters",
                                                             description="Sub-configurations for neurobeam adapters")

    def __str__(self):
        string_to_print = FORMAT_TERMINAL(f"{self.__name__()}\n", "header")
        string_to_print += super().__str__()
        return string_to_print

    @computed_field
    @property
    def locked(self) -> bool:
        """
        :Getter: Whether the configuration is locked and immutable
        :Getter Type: bool
        :Setter: Lock the configuration. Cannot be unlocked.
        :Setter Type: bool
        """
        return self.model_config.get("frozen")

    @computed_field
    @property
    def reporting_interval(self) -> Time:
        """
        :Getter: The interval between progress reports (microseconds)
        :Getter Type: Time
        """
        return Seconds(1 / self.reporting_frequency)

    @computed_field
    @property
    def save_location(self) -> Path:
        """
        :Getter: The path to the save location
        :Getter Type: Path
        """
        return self.save_directory.joinpath(
            f"{self.subject.replace(' ', '_')}_"
            f"{self.study.replace(' ', '_')}_"
            f"{self.condition.replace('-', '_')}_"
            f"{self.date.replace('-', '_')}")

    @computed_field
    @property
    def file_header(self) -> str:
        """
        :Getter: The header for filenames
        :Getter Type: str
        """
        return (f"{self.subject.replace(' ', '_')}_"
                f"{self.study.replace(' ', '_')}_"
                f"{self.condition.replace('-', '_')}_"
                f"{self.date.replace('-', '_')}_"
                )

    @staticmethod
    def __name__() -> str:
        return "Probe Configuration"

    @field_validator("save_directory", "gui_settings", mode="before")
    @classmethod
    def _coerce_path(cls: "Config", v: Union[str, Path]) -> Path:
        """
        Coerce a string into a Path object
        :param v: The string to be coerced
        :returns: The string as a Path object
        """
        return Path(v)

    @locked.setter
    def locked(self, value: bool) -> "Config":
        if not self.model_config["frozen"]:
            self.model_config["frozen"] = value
        else:
            warnings.warn(ImmutableInstanceWarning("Configuration is locked and immutable"), stacklevel=1)

    def add_component(self, config: "ConfigTemplate") -> "Config":
        """
        Add a component configuration to the neurobeam configuration

        :param config: The component configuration to add
        :raises: ImmutableInstanceWarning: Raised if the configuration is locked and immutable
        """
        if self.locked:
            warnings.warn(ImmutableInstanceWarning("Configuration is locked and immutable"), stacklevel=2)
        elif self.components is None:
            self.components = (config,)
        else:
            self.components += (config,)

    def add_adapter(self, config: "ConfigTemplate") -> "Config":
        """
        Add an adapter configuration to the neurobeam configuration

        :param config: The adapter configuration to add
        :raises: ImmutableInstanceWarning: Raised if the configuration is locked and immutable
        """
        if self.locked:
            warnings.warn(ImmutableInstanceWarning("Configuration is locked and immutable"), stacklevel=2)
        elif self.adapters is None:
            self.adapters = (config,)
        else:
            self.adapters += (config,)

    def generate_filename(self, name: str, ext: Optional[str] = None) -> Path:
        """
        Generate a filename for a file related to the current configuration

        :param name: The name of the file
        :param ext: The extension of the file
        """
        if EMPTY(ext) and "." in name:
            name, ext = name.split(".")[-2:]
        if "." not in ext:
            ext = f".{ext}"
        filename = self.save_location.joinpath(f"{self.file_header}{name}").with_suffix(ext)
        return filename

    def save(self, save_location: Optional[Union[str, Path]] = None) -> None:
        """
        Save the configuration to a TOML file

        :param save_location: The location to save the configuration
        """
        if HAS(save_location):
            save_config(save_location, self)
        else:
            save_config(self.generate_filename("configuration", ".toml"), self)

    @model_validator(mode="after")
    def _validate_user_interface(self) -> "Config":
        """
        Validate the user interface settings

        :raises DualUserInterfaceError: Raised if both console and GUI are enabled
        :raises HeadlessWarning: Raised if neither console nor GUI are enabled
        """
        if self.use_console and self.use_gui:
            raise DualUserInterfaceError()
        if not self.use_console and not self.use_gui:
            warnings.warn(HeadlessWarning(), stacklevel=2)
        return self



from importlib_metadata import metadata as _metadata


"""
Metadata of neurobeam installation.
"""


name = _metadata("neurobeam")["name"]
version = _metadata("neurobeam")["version"]
author = _metadata("neurobeam")["author"]
maintainer = _metadata("neurobeam")["maintainer"]



import pytest

from neurobeam.factories import ComponentFactory
from neurobeam.configs import DAQConfig, SpeakerConfig


def test_component_factory(config):
    config.save_location.mkdir(parents=True, exist_ok=True)
    config.add_component(SpeakerConfig())
    config.add_component(DAQConfig())
    factory = ComponentFactory(config)
    components = [factory.registry[component]() for component in factory.registry]
    assert len(components) == len(factory.registry)



from __future__ import annotations
from typing import Tuple, Union, List
from operator import eq
import warnings

from ..exceptions import ImmutableInstanceWarning, SingletonError
from .pattern_matching import PatternMatching


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  COLOR SCHEMES
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class _ColorScheme:
    """
    A container class for neurobeam's color scheme

    .. raises:: :class:`SingletonError` if instance already exists, :class:`ImmutableInstanceWarning` if attribute
        setting is attempted
    """
    BLUE: Tuple[float, float, float] = (15 / 255, 159 / 255, 255 / 255)
    ORANGE: Tuple[float, float, float] = (253 / 255, 174 / 255, 97 / 255)
    GREEN: Tuple[float, float, float] = (64 / 255, 204 / 255, 139 / 255)
    RED: Tuple[float, float, float] = (255 / 255, 78 / 255, 75 / 255)
    PURPLE: Tuple[float, float, float] = (120 / 255, 64 / 255, 204 / 255)
    YELLOW: Tuple[float, float, float] = (255 / 255, 243 / 255, 109 / 255)
    SHADOW: Tuple[float, float, float] = (224 / 255, 224 / 255, 224 / 255)
    LIGHT: Tuple[float, float, float] = (192 / 255, 192 / 255, 192 / 255)
    MEDIUM: Tuple[float, float, float] = (128 / 255, 128 / 255, 128 / 255)
    DARK: Tuple[float, float, float] = (65 / 255, 65 / 255, 65 / 255)
    BLACK: Tuple[float, float, float] = (0 / 255, 0 / 255, 0 / 255)
    WHITE: Tuple[float, float, float] = (255 / 255, 255 / 255, 255 / 255)
    BACKGROUND: Tuple[float, float, float] = (245 / 255, 245 / 255, 245 / 255)
    colors: Tuple[str, ...] = ("red", "green", "blue", "orange", "purple", "yellow", "black", "medium", "dark", "light")

    def __new__(cls: "_ColorScheme") -> "_ColorScheme":
        """
        Force color scheme as singleton!

        """
        if not hasattr(cls, "instance"):
            cls.instance = super(_ColorScheme, cls).__new__(cls)
        else:
            raise SingletonError(cls)
        return cls.instance

    def __setattr__(self, key: Any, value: Any) -> "ColorScheme":
        """
        Prevent setting of attributes

        """
        raise ImmutableInstanceWarning(self)

    @property
    def mapping(self) -> List[Tuple[int, Tuple[float, float, float]]]:
        """
        Mapping of integers to ordered color scheme

        :rtype: :class:`List <typing.List>`\[:class:`Tuple <typing.Tuple>`\[:class:`int`\,
            :class:`Tuple <typing.Tuple>`\[:class:`float`\, :class:`float`\, :class:`float`\]]]
        """
        return list(enumerate(self.colors))

    @property
    def num_colors(self) -> int:
        """
        Total number of colors

        :rtype: :class:`int`
        """
        return len(self.colors)

    def __call__(self, value: Union[int, str], *args, **kwargs) -> Tuple[float, float, float]:
        """
        Call for next color in scheme

        :param value: Requested color directly or by index
        :returns: Tuple of RGB values
        """
        if isinstance(value, int):
            if value >= self.num_colors:
                value -= self.num_colors * (value // self.num_colors)
            return self.mapping[value][1]
        elif isinstance(value, str):
            return getattr(self, value)
        else:
            raise TypeError("color scheme accepts int or str args only")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TERMINAL FORMATTING SCHEME
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class _TerminalScheme:
    """
    A container class for neurobeam's terminal printing color/font scheme

    .. raises:: :class:`SingletonError` if instance already exists, :class:`ImmutableInstanceWarning` if attribute
        setting is attempted
    """
    BLUE: str = "\u001b[38;5;39m"  # Types
    YELLOW: str = "\u001b[38;5;11m"  # Emphasis
    GREEN: str = "\033[38;2;64;204;139m"
    ORANGE: str = "\033[38;2;253;174;97m"
    RED: str = "\033[38;2;255;78;75m"
    BLUE: str = "\033[38;2;15;159;255m"
    BOLD: str = "\u001b[1m"  # Titles, Headers + YELLOW
    UNDERLINE: str = "\u001b[7m"  # Titles, Headers + BOLD + UNDERLINE + YELLOW, implemented as a reverse of font
    # color/background color on some terminals (e.g., PyCharm)
    RESET: str = "\033[0m"

    def __new__(cls: "_TerminalScheme") -> "_TerminalScheme":
        """
        Force color scheme as singleton

        """
        if not hasattr(cls, "instance"):
            cls.instance = super(_TerminalScheme, cls).__new__(cls)
        else:
            raise SingletonError(cls)
        return cls.instance

    def __setattr__(self, key: Any, value: Any):
        """
        Prevent setting of attributes

        """
        raise ImmutableInstanceWarning(self)

    def __str__(self):
        return "Scheme for PROBE terminal printing"

    @property
    def type(self) -> str:
        """
        :Getter: Blue font style for type hinting
        :Getter Type: :class:`str`

        """
        return self.BLUE

    @property
    def emphasis(self) -> str:
        """
        :Getter: Yellow font style for emphasis
        :Getter Type: :class:`str`

        """
        return self.YELLOW

    @property
    def header(self) -> str:
        """
        :Getter: Style for headers, titles and other things of utmost importance consisting of
            bold yellow font and underline (implemented as a reverse of font color / background on some terminals
            (e.g., PyCharm)
        :Getter Type: :class:`str`

        """
        return self.BOLD + self.UNDERLINE + self.YELLOW

    @property
    def announcement(self) -> str:
        return self.BOLD + self.RED + self.UNDERLINE

    @property
    def local_progress(self) -> str:
        return self.GREEN

    @property
    def component(self) -> str:
        return self.BOLD + self.UNDERLINE + self.BLUE

    @property
    def adapter(self) -> str:
        return self.BOLD + self.UNDERLINE + self.GREEN

    @property
    def behavior(self) -> str:
        return self.BOLD + self.UNDERLINE + self.ORANGE

    @property
    def global_progress(self) -> str:
        return self.BLUE

    @staticmethod
    def __name__() -> str:
        return "Terminal Scheme"

    def __repr__(self):
        return "Scheme for PROBE terminal printing"

    def __call__(self, message: str, style: str) -> str:
        """
        Returns properly formatted string without setting global style if appending additional messages

        :param message: string to be formatted
        :param style: desired format (type, emphasis, header or class var)

        :return: formatted string
        """
        # I could probably just have this fail directly, but this is a bit more graceful.
        # It's more important that the message to the user is received than raising an exception because of style
        # matters.

        with PatternMatching(style, eq) as case:
            if any([case(style_) for style_ in dir(self) if "__" not in style_]):
                return "".join([getattr(self, style), message, self.RESET])
            else:
                return "".join([message, f"--requested terminal printing style {style} does not exist"])


# instance persistent color scheme
COLORS = _ColorScheme()

# instance persistent terminal scheme
FORMAT_TERMINAL = _TerminalScheme()



from .camera import Camera
from .daq import DAQ
from .microscope import Microscope
from .slm import SLM
from .speaker import Speaker


# Autoregister all hardware classes
from . import camera
from . import daq
from . import microscope
from . import slm
from . import speaker


from ..registries import ComponentRegistry, auto_register

__modules__ = (camera, daq, microscope, slm, speaker)

for module_ in __modules__:
    auto_register(module_, ComponentRegistry)


# These are the things that should actually be exposed
__all__ = [
    "Camera",
    "DAQ",
    "Microscope",
    "SLM",
    "Speaker",
]



import pytest
import importlib
from inspect import isclass


def test_exceptions(helper):
    probe_exceptions = importlib.import_module("neurobeam.exceptions")
    for exception_ in vars(probe_exceptions).values():
        if isclass(exception_) and issubclass(exception_, Exception):
            with pytest.raises(exception_):
                # noinspection PyArgumentList
                raise exception_(**helper.mock_arguments(exception_))



import pytest
from neurobeam.configs import save_config, load_config


def test_config_io(full_config, helper):
    full_config.save_location.mkdir(parents=True, exist_ok=True)
    path = full_config.generate_filename("test_config", ".toml")
    save_config(path, full_config)
    loaded = load_config(path)
    assert full_config == loaded
    with helper.BlockPrinting():
        print(loaded)



from functools import partial
from typing import Union, Tuple, Any, Optional

from ..extensions import BehaviorState
from ..registries import BehaviorRegistry
from ..tools import HAS


@BehaviorRegistry.register()
def state_transition(origin: "BehaviorState", target: "BehaviorState") -> "BehaviorState":
    # noinspection PyNoneFunctionAssignment
    origin.transition()
    target.initialize()
    return target


@BehaviorRegistry.register()
def ordered_transition(origin: "BehaviorState",
                       targets: Tuple["BehaviorState", ...],
                       order: Tuple[int, ...],
                       final_target: Optional["BehaviorState"] = None) -> "BehaviorState":

    state_map = {idx: state for idx, state in enumerate(targets)}

    if HAS(final_target):
        final_id = len(targets)
        state_map[final_id] = final_target

    def inner_generator():
        return (partial(state_transition, origin=origin, target=state_map[idx]) for idx in order)

    def outer_generator(inner_generator_):
        return (func() for func in inner_generator_)

    def iterator(outer_generator_):
        try:
            return next(outer_generator_)
        except StopIteration:
            raise StopIteration

    return iterator(outer_generator(inner_generator()))



from typing import Tuple  # Optional
from weakref import proxy

from sys import float_info

MAX = float_info.max


from ..configs import HoistedConfig
from ..requests import AdapterRequest
from ..registries import Provider, CallbackRegistry
from ..tools import HAS


@Provider()
class StimulusTrigger:
    def __init__(self,
                 config: "HoistedConfig" = None,
                 name: str = "Stimulus Trigger",
                 component=None,
                 relative_onset: float = 10000.0,
                 ):
        self.config = config
        self.name = config.name if HAS(config) else name
        self.component = component = proxy(component) if component else (
            AdapterRequest(self,
                           "AudioStimulus",
                           alias="component",
                           wildcard=("name", self.config.stimulus_name)))
        self.relative_onset = config.relative_onset if HAS(config) else relative_onset
        self._running = False
        self.onset = MAX
        self._initialized = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        if value and not self._initialized:
            self._initialized = value

    @property
    def running(self):
        return self._running

    @running.setter
    def running(self, value):
        self._running = value

    def command(self):
        self.component.write()

    def read(self):
        return self.running

    @CallbackRegistry.register(alias="stimulus_trigger_trigger")
    def trigger(self, time):
        if self.read() and time >= self.onset:
            self.command()
            self.running = True

    def stop(self):
        self.running = False

    @CallbackRegistry.register(alias="stimulus_trigger_initialize")
    def initialize(self, start_time):
        self.onset = start_time - self.component.component.latency + self.relative_onset
        self.component.put()

    def __call__(self):
        print(f"Triggering stimulus: {self.name}")
        self.command()



from typing import Any
from .static import SUPPORTED_MICROSCOPES


# TODO: Prune unused exceptions
"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// General Errors & Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class UISettingWarning(UserWarning):
    """
    Raised when the configuration and command line arguments for the user interface are inconsistent
    """
    def __init__(self, config_setting: bool, command_setting: bool):
        self.config_setting = config_setting
        self.command_setting = command_setting
        super().__init__(f"The config setting for the user interface is {self.config_setting} while the command line"
                         f"setting is {self.command_setting}. Using the command line setting...")


class OperatingSystemUnsupportedError(OSError):
    """
    Raised when the operating system is not supported
    """
    def __init__(self, os_name: str):
        self.os_name = os_name
        super().__init__(f"Probe is only supported on Windows operating systems. Detected {self.os_name}")


class NotPermittedTypeError(TypeError):
    """
    Raised when a type is not permitted
    """
    def __init__(self, key: str, pos: int, permitted: Any):
        self.key = key
        self.pos = pos
        self.permitted = permitted
        super().__init__(f"Argument {self.key} at position {self.pos} must be of type {self.permitted}")


class SingletonError(RuntimeError):
    """
    Raised when attempting to create a second instance of a singleton
    """

    def __init__(self, singleton: object):
        self.singleton = singleton
        self.name = self.singleton.__name__ if hasattr(self.singleton, "__name__") \
            else type(self.singleton).__name__
        super().__init__(f"{self.name} is a singleton and cannot be instantiated more than once")


class NoWeakRefError(RuntimeError):
    """
    Raised when an object is not able to be weakly referenced
    """
    def __init__(self, obj: object):
        self.obj = obj
        super().__init__(f"Object {self.obj} cannot be weakly referenced")


class DispatchError(RuntimeError):
    """
    Raised when a dispatch method fails
    """
    def __init__(self, obj: object):
        self.obj = obj
        super().__init__(f"Dispatch failed for object {self.obj}")


class ImmutableInstanceWarning(RuntimeWarning):
    """
    Raised when attempting to set an attribute on an immutable instance
    """
    def __init__(self, instance: object):
        self.instance = instance
        super().__init__(f"{self.instance.__class__.__name__} is immutable and cannot be modified")


class LegacyWindowsWarning(UserWarning):
    """
    Raised when using legacy windows
    """
    def __init__(self, os_name: str):
        self.os_name = os_name
        super().__init__(f"Using legacy Windows operating system {self.os_name}."
                         f"Probe's performance may be affected. See 'timer' documentation"
                         f"for details.")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Configuration Errors and Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class UpdateVersionWarning(UserWarning):
    """
    Raised when the configuration version should be updated
    """
    def __init__(self, version: str, expected: str):
        self.version = version
        self.expected = expected
        super().__init__(f"It is advised to update neurobeam to the latest version at your earliest convenience!"
                         f"Configuration version {self.version} does not match the expected version {self.expected}")


class VersionForwardCompatibilityWarning(UserWarning):
    """
    Raised when the configuration version does not match the expected version (forward compatibility of major versions)
    """
    def __init__(self, version: str, expected: str):
        self.version = version
        self.expected = expected
        super().__init__(f"Forward Compatibility of major versions is not guaranteed! "
                         f"Configuration version {self.version} does not match the expected version {self.expected}")


class VersionBackwardCompatibilityWarning(UserWarning):
    """
    Raised when the configuration version does not match the expected version (backward compatibility of minor versions)
    """

    def __init__(self, version: str, expected: str):
        self.version = version
        self.expected = expected
        super().__init__(f"Backwards compatibility of minor versions is not guaranteed!"
                         f"Configuration version {self.version} does not match the expected version {self.expected}")


class ValidationFailureWarning(UserWarning):
    """
    Raised when a configuration update fails due to the value set not matching validation requirements
    """
    def __init__(self, key: str, value: Any):
        self.key = key
        self.value = value
        super().__init__(f"Failed to update configuration key {self.key} with value {self.value}")


class ConfigurationFallbackWarning(RuntimeWarning):
    """
    Raised when falling-back to existing configuration
    """
    def __init__(self):
        super().__init__("Invalid configuration. Falling-back to existing configuration.")


class NoConfigurationWarning(RuntimeWarning):
    """
    Raised when trying to set a sub-configuration when no base configuration
    is available
    """

    def __init__(self, sub_configuration: object):
        self.sub_configuration = sub_configuration
        self.name = sub_configuration.__name__ if hasattr(sub_configuration, "__name__") \
            else type(sub_configuration).__name__
        super().__init__(f"Cannot set {self.name} without a base configuration")


class HeadlessWarning(UserWarning):
    """
    Raised when the user interface is headless
    """
    def __init__(self):
        super().__init__("Running in headless mode. No user interface will be displayed.")


class DualUserInterfaceError(AssertionError):
    """
    Raised when attempting to use both GUI and Console User Interfaces
    """
    def __init__(self):
        super().__init__("Cannot use both GUI and Console User Interfaces")


class VersionBackwardCompatibilityError(ValueError):
    """
    Raised when the configuration version does not match the expected version
    """
    def __init__(self, version: str, expected: str):
        self.version = version
        self.expected = expected
        super().__init__(f"Probe does not support backwards compatibility of major versions! Configuration version "
                         f"{self.version} does not match the expected version {self.expected}")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Factory or Registry Errors and Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class DuplicateRegistrationError(RuntimeError):
    """
    Raised when the factory fails to register a class due to duplicate keys
    """
    def __init__(self, factory: object, class_to_register: object):
        self.factory = factory
        self.class_to_register = class_to_register
        super().__init__(f"Factory {self.factory.__class__.__name__} failed to register {self.class_to_register} "
                         f"due to a previously registered constructor with an identical key")


class MissingIdentifierError(KeyError):
    """
    Raised when an identifier is missing
    """
    def __init__(self, factory: object, identifier: str):
        self.factory = factory
        self.identifier = identifier
        super().__init__(f"Identifier {self.identifier} is missing in {self.factory.__class__.__name__}")


class InvalidComponentError(AttributeError):
    """
    Raised when an invalid component is registered
    """
    def __init__(self, component: object):
        self.component = component
        super().__init__(f"Component {self.component} is not a valid neurobeam component.")


class InvalidAdapterError(AttributeError):
    """
    Raised when an invalid adapter is registered
    """
    def __init__(self, adapter: object):
        self.adapter = adapter
        super().__init__(f"Adapter {self.adapter} is not a valid neurobeam adapter.")


class InvalidAudioError(AttributeError):
    """
    Raised when an invalid audio is registered
    """
    def __init__(self, audio: object):
        self.audio = audio
        super().__init__(f"Audio {self.audio} is not a valid neurobeam audio.")


class InvalidCallbackError(AttributeError):
    """
    Raised when an invalid callback is registered
    """
    def __init__(self, callback: object):
        self.callback = callback
        super().__init__(f"Callback {self.callback} is not a valid neurobeam callback.")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Controller Errors and Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class ControllerImmutabilityWarning(RuntimeWarning):
    """
    Raised when attempting to modify the controller when running & immutable
    """
    def __init__(self):
        super().__init__("Controller is running and immutable. Cannot modify components or adapters.")


class EmptyControllerWarning(RuntimeWarning):
    """
    Raised when the controller is empty
    """
    def __init__(self):
        super().__init__("Controller is empty. Please add components and adapters.")


class EmptyControllerError(RuntimeError):
    """
    Raised when the controller is empty
    """
    def __init__(self):
        super().__init__("Controller is empty. Cannot execute control methods without components and "
                         "adapters to control.")


class InvalidComponentWarning(RuntimeWarning):
    """
    Raised when an invalid object is passed to the controller as a positional argument or through the
    :method:`Controller.add_component <neurobeam.controllers.Controller.add_component>` method
    """
    def __init__(self, component: Any):
        self.component = component
        super().__init__(f"Component {self.component} is not a valid neurobeam component.")


class InvalidAdapterWarning(RuntimeWarning):
    """
    Raised when an invalid adapter is passed to the controller as a positional argument or through the
    :method:`Controller.add_adapter <neurobeam.controllers.Controller.add_adapter>` method
    """
    def __init__(self, adapter: Any):
        self.adapter = adapter
        super().__init__(f"Adapter {self.adapter} is not a valid neurobeam adapter.")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Filesystems / Data Buffers Errors and Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class DataReorganizationWarning(RuntimeWarning):
    """
    Raised when data reorganization was unsuccessful
    """
    def __init__(self, elements: int, shape: tuple):
        self.elements = elements
        self.shape = shape
        super().__init__(f"The number of elements {self.elements} in the written data cannot be coerced to the "
                         f"requested shape {self.shape}")


class DesynchronizedBuffersWarning(RuntimeWarning):
    """
    Raised when a group of buffers is desynchronized
    """
    def __init__(self, buffers: tuple):
        self.buffers = buffers
        super().__init__(f"The following buffers are desynchronized: {self.buffers}")


class EmptyEventLogWarning(RuntimeWarning):
    """
    Raised when argument to the event log is empty
    """
    def __init__(self):
        super().__init__("Event log was not supplied any argument")


class InvalidExtensionWarning(UserWarning):
    """
    Raised when an invalid file extension is used
    """
    def __init__(self, key: str, pos: int, extension: str, permitted: Any):
        self.key = key
        self.pos = pos
        self.extension = extension
        self.permitted = permitted
        super().__init__(f"Argument {self.key} at position {self.pos} has invalid file extension {self.extension}. "
                         f"Expected extension {self.permitted} and coerced to {self.permitted}.")


class OverloadedEventLogWarning(RuntimeWarning):
    """
    Raised when the event log is overloaded
    """
    def __init__(self, args: Any):
        self.args = args
        super().__init__(f"Event log was overloaded with {len(self.args)} arguments: {self.args}")


class PathMismatchWarning(RuntimeWarning):
    """
    Raised the eventlog is successfully called from the instance registry but a path is provided that does not match
    the path located in the instance registry.
    """
    def __init__(self, key: str, path: str, expected: str):
        self.key = key
        self.path = path
        self.expected = expected
        super().__init__(f"The passed instance registry key {self.key} has path {self.expected} that does not match the"
                         f"passed path {self.path}")


class RingSizeMismatchWarning(RuntimeWarning):
    """
    Raised the eventlog is successfully called from the instance registry but a ring size is provided that does not
    match the ring size specified by the existing instance.
    """

    def __init__(self, key: str, ring_size: int, expected: int):
        self.key = key
        self.ring_size = ring_size
        self.expected = expected
        super().__init__(
            f"The passed instance registry key {self.key} has ring size {self.expected} that does not match the"
            f"passed ring size {self.ring_size}")


class InvalidFilenameError(ValueError):
    """
    Raised when an invalid filename is used
    """
    def __init__(self, key: str, pos: int, filename: str):
        self.key = key
        self.filename = filename
        self.pos = pos
        super().__init__(f"Argument {self.key} at position {self.pos} has invalid filename {self.filename}."
                         f"Please use only alphanumeric characters and underscores.")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Callback Errors and Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class CallbackError(RuntimeError):
    """
    Raised when a component calls a callback that has not been injected

    """
    def __init__(self, obj: str, callback: str):
        self.obj = obj
        self.callback = callback
        super().__init__(f"Component {self.obj} called {self.callback}, but it was not injected!")


class AmbiguousCallbackInjectionError(RuntimeError):
    """
    Raised when the injector fails to inject an approximate callback into a component due to multiple callbacks
    matching the search tag.

    """
    def __init__(self, obj: str, callback: str):
        self.obj = obj
        self.callback = callback
        super().__init__(f"Callback with tag {self.callback} is ambiguous for {self.obj}; multiple callbacks "
                         f"match the search tag.")


class SingularCallbackCollectionError(RuntimeError):
    """
    Raised when the injector is able to locate only one callback within a required callback collection request.
    """
    def __init__(self, obj: str, callback: str):
        self.obj = obj
        self.callback = callback
        super().__init__(f"Component {self.obj} requested a required callback collection with tag "
                         f"{self.callback} but only one callback was located.")


class RequiredCallbackCollectionError(RuntimeError):
    """
    Raised when the injector fails to inject a required callback collection into a component
    """
    def __init__(self, obj: str, callback: str):
        self.obj = obj
        self.callback = callback
        super().__init__(f"Failed to inject required callback collection {self.callback} into {self.obj}")


class RequiredCallbackInjectionError(RuntimeError):
    """
    Raised when a component fails to receive a required callback
    """
    def __init__(self, obj: str, callback: str):
        self.obj = obj
        self.callback = callback
        super().__init__(f"Failed to inject required callback {self.callback} into {self.obj}")


class InjectionError(RuntimeError):
    """
    Raised when the controller fails to inject or a component fails
    to receive a required callback.
    """
    def __init__(self, component: str, callback: str):
        self.component = component
        self.callback = callback
        super().__init__(f"Failed to inject {self.callback} into {self.component}")


class NoOptionalCallbackWarning(UserWarning):
    """
    Raised when an optional callback was not available
    """
    def __init__(self, component: str, callback: str):
        self.component = component
        self.callback = callback
        super().__init__(f"Component {self.component} requested an optional callback with tag {self.callback} but it"
                         f" was not located.")


class NoOptionalCallbackCollectionWarning(UserWarning):
    """
    Raised when an optional callback collection was not available
    """
    def __init__(self, component: str, callback: str):
        self.component = component
        self.callback = callback
        super().__init__(f"Component {self.component} requested an optional callback collection with tag \
        {self.callback} but it was not located.")


class SingularCallbackCollectionWarning(RuntimeWarning):
    """
    Raised when the injector is able to locate only one callback within an optional or approximate
    callback collection request.
    """
    def __init__(self, component: str, callback: str):
        self.component = component
        self.callback = callback
        super().__init__(f"Component {self.component} requested an approximate or optional callback collection with"
                         f" tag {self.callback} but only one callback was located.")


class CallbackKeyWarning(RuntimeWarning):
    """
    Raised when a callback key is not present in the requesting instance
    """
    def __init__(self, instance: Any, key: str):
        self.instance = instance
        self.key = key
        super().__init__(f"Callback key {self.key} is not present in the requesting instance {self.instance}")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Behavior Errors and Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class InvalidStateError(RuntimeError):
    """
    Raised when the behavioral state machine or another object encounters an invalid state
    """
    def __init__(self, name: str):
        self.state = name
        super().__init__(f"State {self.state} is invalid")


class InvalidStateEntryError(RuntimeError):
    """
    Raised when the behavioral state machine attempts to enter a state in a manner that cannot possibly be valid
    """
    def __init__(self, name: str):
        self.state = name
        super().__init__(f"State {self.state} cannot be entered")


class InvalidTransitionError(RuntimeError):
    """
    Raised when the behavioral state machine attempts to transition to a state in a manner that cannot possibly be valid
    """
    def __init__(self, name: str):
        self.transition = name
        super().__init__(f"Transition {self.transition} is invalid")


class InvalidBehavioralDesignWarning(RuntimeWarning):
    """
    Raised when the behavioral design is invalid
    """
    def __init__(self):
        super().__init__("Behavioral design is invalid. Please check that each state is connected to at least "
                         "one transition.")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Microscope Errors and Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class UnsupportedMicroscopeError(RuntimeError):
    """
    Raised when an unsupported microscope is used
    """
    def __init__(self, microscope: str):
        self.microscope = microscope
        super().__init__(f"Microscope {self.microscope} is not supported. "
                         f"Supported microscopes are {SUPPORTED_MICROSCOPES}")


class MicroscopeConstructionError(RuntimeError):
    """
    Raised when no microscopes were constructed
    """


class MicroscopeConnectionError(ConnectionError):
    """
    Raised when a microscope connection fails
    """


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Video / Camera Errors and Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class CameraConstructionError(RuntimeError):
    """
    Raised when cameras were constructed
    """
    def __init__(self, camera_idx: int):
        self.camera_idx = camera_idx
        super().__init__(f"The following cameras were not constructed: {self.camera_idx}")


class CameraCollectionWarning(RuntimeWarning):
    """
    Raised when camera fails to read a frame

    """
    def __init__(self, camera_idx: str):
        self.camera_idx = camera_idx
        super().__init__(f"Failed to read frame from camera {self.camera_idx}")


class NoCamerasWarning(UserWarning):
    """
    Raised when no cameras were included in video configuration
    """
    def __init__(self):
        super().__init__("If 'use_video' is true in VideoConfigTemplate, "
                         "then at least one camera must be included.")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// National Instruments DAQ Errors & Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Speaker Errors & Warnings
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class SpeakerSearchError(KeyError):
    """
    Raised when speaker was not constructed
    """
    def __init__(self, speaker_id: str, found_speakers: dict):
        self.speaker_id = speaker_id
        super().__init__(f"Speaker was not found using the following search id: {self.speaker_id}. These speakers"
                         f" are available {found_speakers.keys()}")


class SpeakerPropertyError(ValueError):
    """
    Raised when selected speaker has an invalid property
    """
    def __init__(self, speaker_id: str, property_name: str, property_value: Any, expected_value: Any):
        self.speaker_id = speaker_id
        self.property_name = property_name
        self.property_value = property_value
        self.expected_value = expected_value
        super().__init__(f"Speaker {self.speaker_id} has an invalid property {self.property_name}. "
                         f"The property must {self.property_value} is not supported: expected {self.expected_value}")


class SpeakerVolumeWarning(RuntimeWarning):
    """
    Raised when speaker volume is out of range
    """
    def __init__(self, request: float, min_request: float, max_request: float, set_volume: float):
        self.request = request
        self.min_request = min_request
        self.max_request = max_request
        self.set_volume = set_volume
        super().__init__(f"Speaker volume request {self.request} decibels generates out of range volume. Minimum volume"
                         f" is {self.min_request} decibels and maximum volume is {self.max_request} decibels. "
                         f"setting to {self.set_volume} decibels instead")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TIME ERRORS & WARNINGS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class TimeUnitParsingWarning(UserWarning):
    """
    Raised when the time unit is not recognized
    """
    def __init__(self, unit: str):
        self.unit = unit
        super().__init__(f"Time unit {self.unit} is not recognized. Defaulting to microseconds.")


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CONTAINER ERRORS & WARNINGS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class ItemNotAddedWarning(UserWarning):
    """
    Raised when an item is not added to the container
    """
    def __init__(self, item: object):
        self.item = item
        super().__init__(f"Item {self.item} was not added to the container")



import webbrowser


def open_contact() -> None:
    """
    Opens the contact page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_documentation() -> None:
    """
    Opens the documentation on read-the-docs for PROBE
    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_donate() -> None:
    """
    Opens the donation page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_examples() -> None:
    """
    Opens the examples page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_github() -> None:
    """
    Opens the GitHub repository for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_issues() -> None:
    """
    Opens the issues page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")


def open_pypi() -> None:
    """
    Opens the PyPi page for PROBE

    """
    webbrowser.open("https://github.com//darikoneil//PROBE")



import pytest

from neurobeam.extensions import CallbackFunction
from neurobeam.requests import bind_callback, fake_callback, rename_callback, wrap_callback_collection, BoundFunction
from neurobeam.requests import _SingleInjection, _CollectionInjection, CallbackRequest, AdapterRequest


def test_bind_callback():
    ...


def test_fake_callback():
    function = fake_callback("faker")
    assert function.__name__ == "faker"
    assert function() is None
    assert isinstance(function, CallbackFunction)


def test_rename_callback():
    function = fake_callback("faker")
    assert function.__name__ == "faker"
    function = rename_callback("renamed")(function)
    assert function.__name__ == "renamed"
    assert function() is None
    assert isinstance(function, CallbackFunction)


def test_wrap_callback_collection():
    function0 = fake_callback("faker0")
    function1 = fake_callback("faker1")
    function2 = fake_callback("faker2")

    wrapped_function = wrap_callback_collection((function0, function1, function2), name="wrapped_function")
    assert wrapped_function.__name__ == "wrapped_function"
    assert wrapped_function() is None
    assert isinstance(wrapped_function, CallbackFunction)


def test_bound_function():
    ...


def test_implementation_selection(MockComponent):
    mc = MockComponent()

    setattr(mc, "mock_single_callback_request", None)
    mc.mock_single_callback_request = CallbackRequest(mc,
                                                      key="reporting_console_start",
                                                      alias="mock_single_callback_request")
    setattr(mc, "mock_collection_callback_request", None)
    mc.mock_collection_callback_request = CallbackRequest(mc,
                                                          ("reporting_console_start", "reporting_console_stop"),
                                                          alias="mock_collection_callback_request")

    assert isinstance(mc.mock_single_callback_request._implementation, _SingleInjection)
    assert isinstance(mc.mock_collection_callback_request._implementation, _CollectionInjection)



from functools import partial
from sys import float_info as _float_info
from threading import Thread
from time import sleep
from typing import Callable, Union

from frozendict import frozendict
import numpy as np

from ..registries import BehaviorRegistry, Provider
from ..requests import CallbackRequest
from ..timing import timer

# Constants
MIN = _float_info.min
MAX = _float_info.max


@BehaviorRegistry.register()
@Provider()
class Machine(Thread):
    """
    Behavioral state machine
    """
    def __init__(self,
                 states: dict[str, "BehaviorState"],
                 transitions: dict[str, Callable],
                 event_log: "EventLog",
                 reporting_interval: float):
        ################################################################################################################
        # INITIALIZE
        ################################################################################################################
        #: Union[dict[str, "BehaviorState"], frozendict[str, "BehaviorState"]]: states
        self.states = states
        #: Union[dict[str, partial], frozendict[str, partial]]: transitions
        self.transitions = transitions
        #: float: reporting interval (microseconds)
        self.reporting_interval = reporting_interval
        #: bool: running flag
        self.running = False
        #: float: next time to report
        self.next_report_time = MAX
        #: float: progress of state
        self.state_progress = 0.0
        #: float: progress of task
        self.task_progress = 0.0
        #: bool: whether the behavioral state machine is locked into its build
        self._locked = False
        self._current_state = None

        ################################################################################################################
        # Request Callbacks
        ################################################################################################################
        #: start_callback: Callable: start callback
        self.startup = CallbackRequest(self, key="reporting_console_task_start", alias="startup")

        #: end_callback: Callable: end callback
        self.end_experiment = CallbackRequest(self, key="destroy", alias="end_experiment")

        #: Tuple[Callable]: reporting callbacks
        self.reporting_progress_state = CallbackRequest(self, key="reporting_console_progress_state",
                                                        alias="reporting_progress_state",
                                                        )
        self.reporting_progress_task = CallbackRequest(self, key="reporting_console_progress_task",
                                                       alias="reporting_progress_task",
                                                       )

        ################################################################################################################
        # PRE-ALLOCATE
        ################################################################################################################
        self.current_state = self.states.get("Setup")

        ################################################################################################################
        # Instantiate Thread
        ################################################################################################################
        Thread.__init__(self, name="BehavioralMachine")

    @property
    def locked(self) -> bool:
        return self._locked

    @property
    def current_state(self) -> "BehaviorState":
        return self._current_state.name

    @locked.setter
    def locked(self, value: bool) -> None:
        if value and not self._locked:
            self.states = frozendict(self.states)
            self.transitions = frozendict(self.transitions) if self.transitions else frozendict()
            self._locked = value
        elif not value and self._locked:
            raise TypeError
        # TODO: import correct error

    @current_state.setter
    def current_state(self, state: "BehaviorState") -> "_BehavioralStateMachine":
        self._current_state = state

    def wrap_all_callbacks(self) -> "_BehavioralStateMachine":
        for state in self.states.values():
            state.wrap_callbacks()

    def start(self) -> "_BehavioralStateMachine":
        self.wrap_all_callbacks()
        self.locked = True
        self._current_state.initialize()
        self.running = True
        self.startup()
        self.next_report_time = timer() + self.reporting_interval
        self.current_state = self.transitions.get("Setup")()
        Thread.start(self)  # Call Thread start method

    def stop(self) -> "_BehavioralStateMachine":
        self.running = False

    def run(self) -> "_BehavioralStateMachine":
        while self.running:
            time = timer()
            if self._current_state.state_loop(time):
                self.current_state = self.transitions.get(self.current_state)()
            self.update_progress()
            self.check_reporting(time)
            self.release_gil()
            if self.current_state == "End":
                self.running = False
                self.end_experiment()
                break

    def check_reporting(self, time: float) -> "_BehavioralStateMachine":
        if time >= self.next_report_time:
            self.reporting_progress_state(self.current_state, self.state_progress)
            self.reporting_progress_task("DARIK", self.task_progress)
            self.next_report_time += self.reporting_interval

    def release_gil(self) -> "_BehavioralStateMachine":
        # Use this to make sure the behavioral thread isn't clogging up the flow running excessive timer checks
        # Python 3.11 added a "high-resolution timer" that reports in the nanoseconds but when checking the polling
        # performance seems to be limited to 1/2 millisecond, and I'd like to avoid having to backport. If the argument
        # to time.sleep() is zero, you will pass the GIL instead which will hopefully avoid this light behavior hogging
        # the instruction list.

        # Update: I need to set sleep to non-zero or the GIL is still clogged...
        sleep(MIN)
        # pass

    def update_progress(self, *args, **kwargs) -> "_BehavioralStateMachine":
        self.state_progress = self._current_state.state_progress
        self.task_progress = self._check_task_progress(*args, **kwargs)

    def _check_task_progress(self, *args, **kwargs) -> float:
        return 0.0



from functools import partial, singledispatchmethod
from typing import Union, Tuple, Optional, Callable, List
from .configs import Config, HoistedConfig, StateConfig, TransitionConfig
from .controller import Controller
from .data import EventLog
from .exceptions import MissingIdentifierError
from .extensions import Component, Adapter, BehaviorState
from .registries import AdapterRegistry, AudioRegistry, BehaviorRegistry, ComponentRegistry
from .tools import prune_keys, parse_empties, HAS, EMPTY


"""
Here are the classes and functions that are used to create the various components of the neurobeam. These factories are used
to create the adapters, components, and controller. Configs are created separating using the Config class. The GUI 
factory is located elsewhere (GUI FOLDER)
"""


class AdapterFactory:
    #: AdapterRegistry: The adapter registry to retrieve adapter constructors from
    _registry = AdapterRegistry()

    def __init__(self, config: "Config"):
        #: Config: The neurobeam configuration
        self.config = config

        #: dict: instance registry of hardware constructors with hoisted configs attached using partial functions
        self.registry = {}

        # Hoist sub-configurations to top-level namespace
        self.hoist()

    def create(self, adapter: Union[str, int]) -> "Adapter":
        if adapter not in self.registry:
            raise MissingIdentifierError(self, adapter)
        return self.registry.get(adapter)()

    def hoist(self) -> "AdapterFactory":
        """
        Hoist the factory methods to the instance registry
        """
        self.registry = \
            {idx: partial(self._registry.get(adapter.factory_key),
                          HoistedConfig(self.config, "adapters", idx))
             for idx, adapter in enumerate(self.config.adapters)}


class AudioFactory:
    #: dict: registry of audio stimulus constructors
    _registry = AudioRegistry()

    def __init__(self, speaker: "Speaker"):
        self._speaker = speaker
        self.registry = {method: partial(self._registry.get(method),
                                         self._speaker.sampling_rate,
                                         self._speaker.channels
                                         ) for method in self._registry}

    @classmethod
    def register(cls):  # noqa: ANN206
        """
        A decorator to register a method as a constructor for a specific audio stimulus type.

        """
        def register_method(method):  # noqa: ANN206, ANN001, ANN201
            cls._registry[method.__name__] = method
            return method
        return register_method

    def create(self, method: str, *args, **kwargs):
        """
        Create an audio stimulus using the specified method.

        Args:
            method (str): The method to use to create the stimulus.
            *args: The arguments to pass to the constructor.
            **kwargs: The keyword arguments to pass to the constructor.

        Returns:
            np.ndarray: The stimulus.

        """
        if "intensity" in kwargs:
            intensities = kwargs.pop("intensity")
            if isinstance(intensities, (tuple, list)):
                kwargs["volume"] = tuple([self._speaker.lut.get(intensity) for intensity in intensities])
            else:
                kwargs["volume"] = self._speaker.lut.get(intensities)
        stimulus = self.registry.get(method)(*args, **kwargs)

        return stimulus


class BehaviorFactory:
    """
    Factory for creating states and transitions for a behavioral state machine. We use a factory here to allow for
    flexibility in the creation of states and transitions, and to allow for the possibility of users creating their own
    states and transitions. This is a singleton class, as we only need one factory and we only ever have one behavioral
    state machine.
    """
    _registry = BehaviorRegistry()

    @staticmethod
    def format_state_recipe(states: Tuple[StateConfig, ...]) -> Tuple[dict, ...]:
        # format
        state_recipe = [dict(state) for state in states]
        # ensure there is a setup
        if not any([state.get("special") == "Setup" for state in state_recipe]):
            state_recipe.append({"special": "Setup"})
        # ensure there is an end
        if not any([state.get("special") == "End" for state in state_recipe]):
            state_recipe.append({"special": "End"})
        return tuple(state_recipe)

    @staticmethod
    def format_transition_recipe(transitions: Tuple[TransitionConfig, ...]) -> Tuple[dict, ...]:
        transition_recipe = [transition.dict() for transition in transitions]
        transition_recipe = tuple(transition_recipe)
        return transition_recipe

    @classmethod
    def construct(cls,
                  states: Tuple[StateConfig, ...],
                  transitions: [TransitionConfig, ...],
                  event_log: "EventLog",
                  reporting_interval: float) -> "_BehaviorFactory":
        # First construct states, then construct transitions
        # (attaching transitions to states through partials as necessary)
        state_recipe = cls.format_state_recipe(states)
        states = {state.name: state for state in cls.construct_states(state_recipe, event_log)}
        # Fail on single state, but we will always have a setup and end state... so it's ok
        transition_recipe = cls.format_transition_recipe(transitions)
        transitions = cls.construct_transitions(transition_recipe, states)
        # if isinstance(transitions, tuple) and len(transitions) == 2:
        #    transitions = {transitions[0]: transitions[1]}
        # Construct & initialize state machine
        machine = cls._registry.get("Machine")
        return machine(states, transitions, event_log, reporting_interval)

    # noinspection PyNestedDecorators
    @singledispatchmethod
    @classmethod
    def construct_states(cls,
                         state_recipe,  # noqa: U100
                         event_log: Optional["EventLog"] = None  # noqa: U100
                         ) -> Union["BehaviorState", List["BehaviorState"]]:  # noqa: U100
        raise TypeError

    # noinspection PyMethodParameters,PyNestedDecorators
    @construct_states.register
    @classmethod
    def _(cls,
          state_recipe: tuple,
          event_log: Optional["EventLog"] = None
          ) -> List["BehaviorState"]:
        return [cls.construct_states(recipe, event_log) for recipe in state_recipe]

    # noinspection PyMethodParameters,PyNestedDecorators
    @construct_states.register
    @classmethod
    def _(cls, state_recipe: dict, event_log: Optional["EventLog"] = None) -> "BehaviorState":
        state_type = state_recipe.get("special")
        if EMPTY(state_type):
            state_type = "State"
        if not cls._registry.has(state_type):
            raise MissingIdentifierError(cls, state_type)
        method = cls._registry.get(state_type)
        state_recipe = prune_keys(state_recipe, ("special", "factory_key", "version"))
        # TODO: necessary?
        state_recipe = parse_empties(state_recipe)
        # TODO: necessary?
        options = state_recipe.get("options", {})
        return method(event_log=event_log, **{**options, **state_recipe})

    # noinspection PyNestedDecorators
    @singledispatchmethod
    @classmethod
    def construct_transitions(cls,
                              transition_recipe: dict) -> "_BehavioralStateMachine":  # noqa: U100
        raise TypeError

    # noinspection PyNestedDecorators
    @construct_transitions.register
    @classmethod
    def _(cls, transition_recipe: tuple, states: dict[str, "BehaviorState"]) -> dict[str, Callable]:
        return dict(cls.construct_transitions(recipe, states) for recipe in transition_recipe)

    # noinspection PyNestedDecorators
    @construct_transitions.register
    @classmethod
    def _(cls, transition_recipe: dict, states: dict[str, "BehaviorState"]) -> Tuple[str, Callable]:
        name = transition_recipe.get("name")
        origin = transition_recipe.get("origin")
        target = transition_recipe.get("targets")
        name = name if name.lower() != "auto" else origin
        if isinstance(target, tuple):
            targets = tuple([states.get(t) for t in target])
            order = transition_recipe.get("order")
            return name, partial(cls._registry.get("ordered_transition"), states.get(origin), targets, order)
        else:
            targets = states.get(target)
            return name, partial(cls._registry.get("state_transition"), states.get(origin), targets)
# TODO: Clean up & debug


class ComponentFactory:
    #: dict: ComponentRegistry: The component registry to retrieve component constructors from
    _registry = ComponentRegistry()

    def __init__(self, config: "Config"):
        """
        Factory for creating components, as defined by :class:`Component <neurobeam.extensions.Component>`. Components are
        added to the registry using the :meth:`register` decorator.

        :raises DuplicateRegistrationError: Raised if the registration cannot be complete on a class because it would
        create duplicate registry keys.
        :raises MissingIdentifierError: Raised if the requested factory method is not in the registry.

        ..seealso:: :class:`Component <neurobeam.extensions.Component>`
        """
        #: Config: The neurobeam configuration
        self.config = config

        #: dict: instance registry of hardware constructors with hoisted configs attached using partial functions
        self.registry = {}

        # Hoist sub-configurations to top-level namespace
        self.hoist()

    def create(self, component: Union[str, int]) -> "Component":
        """
        Create a component using the specified constructor. Retrieves a partial function from the instance registry to
        automatically pass the config to the constructor.

        :param component: The factory_key or the index of the component to create
        :returns: The instantiated component.
        :raises MissingIdentifierError: Raised if the requested factory method is not in the registry.
        """
        if component not in self.registry:
            raise MissingIdentifierError(self, component)
        return self.registry.get(component)()

    def hoist(self) -> "ComponentFactory":
        """
        Hoist the factory methods to the instance registry
        """
        self.registry = \
            {idx: partial(self._registry.get(component.factory_key),
                          HoistedConfig(self.config, "components", idx))
             for idx, component in enumerate(self.config.components)}



import pytest

from neurobeam.timing import timer, Microseconds, Milliseconds, Seconds, Minutes, Hours

import numpy as np


COVERAGE_OVERHEAD = 2.5  # overhead of running tests with coverage, which impacts timer performance


def test_timer():

    deltas = np.diff([timer() for time_step in range(25)])
    # Down from 1000 because for some reason we have large delays when using coverage
    assert (np.median(deltas) <= 2 * COVERAGE_OVERHEAD), f"Poor timer performance: median time resolution = " \
                                                         f"{np.median(deltas)}"
    assert (np.max(deltas) <= 15 * COVERAGE_OVERHEAD), f"Poor timer performance: worst case time resolution = " \
                                                       f"{np.max(deltas)}"


def test_conversions():

    test_time_hours = 1.25 / 3600
    test_time_minutes = 1.25 / 60
    test_time_seconds = 1.25
    test_time_milli = 1250
    test_time_micros = 1250000

    times = (test_time_micros, test_time_milli, test_time_seconds, test_time_minutes, test_time_hours)

    # Test base
    base_time = Microseconds(test_time_micros)
    attrs = ("microseconds", "milliseconds", "seconds", "minutes", "hours")
    for attr, unit_of_time in zip(attrs, times):
        assert (unit_of_time == getattr(base_time, attr))

    # test converters

    converters = (Microseconds, Milliseconds, Seconds, Minutes, Hours)
    for converter, unit_of_time in zip(converters, times):
        assert (test_time_micros == converter(unit_of_time))



from functools import partial
from itertools import count


class _CallableCounter(count):
    def __call__(self):
        return next(self)


def _inner_counter(_counter: _CallableCounter):
    return _counter()


def counter() -> partial:
    _counter = _CallableCounter()
    return partial(_inner_counter, _counter)



from ctypes import c_int64, byref, windll
from typing import Any, Union
import warnings

from .exceptions import TimeUnitParsingWarning


"""
Classes and functions related to timing and time representation.
"""


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// HIGH-PERFORMANCE TIMER
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


# this binding retrieves the current number of ticks from the windows performance counter
_k32_cnt = windll.Kernel32.QueryPerformanceCounter
# this binding retrieves the frequency of ticks collected by the windows performance counter
_k32_freq = windll.Kernel32.QueryPerformanceFrequency


def timer(units: int = 10**6) -> float:
    """
    Glue function for high-performance timer. Calls the windows performance counter and frequency through C-bindings.
    Note that these functions are not available on all platforms, and require that the processor's time-stamp counter
    is invariant across cores and threads, as well as invariant through boosting and changes in power-state. The vast
    majority of modern processors satisfy these requirements, and in the even they do not, modern windows versions
    (i.e., Windows 8/10/11) will automatically fal-lback to an invariant clock. This fal-lback clock is not as
    high-performance and high-resolution as the query performance counter, so you will end up with considerable larger
    intervals between timestamps and less precision in timing. While these effects are not relevant in most
    applications, it becomes very relevant in real-time systems like neurobeam.

    :param units: The units to return the time in. Default is microseconds.
    :returns: The relative time in the units of microseconds.

    .. seealso:: https://learn.microsoft.com/en-us/windows/win32/api/profileapi/nf-profileapi-queryperformancefrequency

    .. seealso:: https://learn.microsoft.com/en-us/windows/win32/api/profileapi/nf-profileapi-queryperformancecounter
    """
    count = c_int64()
    frequency = c_int64()

    _k32_cnt(byref(count))
    _k32_freq(byref(frequency))

    return count.value * units / frequency.value


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TIME REPRESENTATION
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class Time(int):
    """
    A class for representing time in various units. The class is a subclass of int, and can be used as such. The class
    will operate in units of microseconds, but by explicitly calling the properties, the time can be converted to
    the desired unit. Note that these conversions return floats, not an integer or a copy of the class.
    """
    _MICROS_PER_MILLISECOND = 10 ** 3
    _MICROS_PER_SECOND = 10 ** 6
    _MICROS_PER_MINUTE = 6 * 10 ** 7
    _MICROS_PER_HOUR = 36 * 10 ** 8

    def __new__(cls, time: int) -> "Time":
        return int.__new__(cls, time)

    @property
    def microseconds(self) -> float:
        """
        Converts the time to microseconds.

        :returns: The time in microseconds
        """
        return self

    @property
    def milliseconds(self) -> float:
        """
        Converts the time to milliseconds

        :returns: The time in milliseconds
        """
        return self / float(self._MICROS_PER_MILLISECOND)

    @property
    def seconds(self) -> float:
        """
        Converts the time to seconds

        :returns: The time in seconds
        """
        return self / float(self._MICROS_PER_SECOND)

    @property
    def minutes(self) -> float:
        """
        Converts the time to minutes

        :returns: The time in minutes
        """
        return self / float(self._MICROS_PER_MINUTE)

    @property
    def hours(self) -> float:
        """
        Converts the time to hours

        :returns: The time in hours
        """
        return self / float(self._MICROS_PER_HOUR)

    @staticmethod
    def __name__() -> str:
        return "Time"

    @classmethod
    def __toml_decode__(cls, data: Union[list, dict]) -> "Time":
        """
        Decodes a Time object from a serialized representation.

        :param data: The serialized representation of the Time object.
        :returns: An instance of Time
        """
        if isinstance(data, dict):
            value, units = data.get("value"), data.get("units")
        else:
            value, units = data
        if units in ("us", "µs", "microseconds"):
            return Microseconds(value)
        elif units in ("m", "minutes"):
            return Minutes(value)
        elif units in ("ms", "milliseconds"):
            return Milliseconds(value)
        elif units in ("s", "seconds"):
            return Seconds(value)
        elif units in ("h", "hours"):
            return Hours(value)
        else:
            warnings.warn(TimeUnitParsingWarning(units), stacklevel=2)
            return Microseconds(value)

    def __toml_encode__(self) -> list:
        """
        Encodes the Time object into a serializable representation.

        :returns: Serialized representation of the Time object.
        """
        name = self.__class__.__name__
        if name == "Time":
            return [self.microseconds, "microseconds"]
        else:
            return [getattr(self, name.lower()), name.lower()]

    def __copy__(self) -> "Time":
        """
        Returns a shallow copy of the Time object.
        """
        return Time(self.microseconds)

    def __deepcopy__(self, memo: dict) -> "Time":
        """
        Returns a deep copy of the Time object.
        """
        return Time(self.microseconds)

    def __instancecheck__(self, instance: Any) -> bool:
        """
        Checks if the instance is an instance of Time or any of its subclasses.

        :param instance: Existing instance to check.
        :returns: Whether the instance is an instance of Time or any of its subclasses.
        """
        if isinstance(instance, (Time, Microseconds, Milliseconds, Seconds, Minutes, Hours)):
            return True
        else:
            return super().__instancecheck__(instance)


class Microseconds(Time):
    """
    A class for representing time in microseconds. The class is a subclass of Time, and the value passed in the
    constructor argument will remain in units of microseconds when constructing an instance of
    :class:`Time <RF_PROBE.timing.Time>`.

    :param microseconds: The time in microseconds.
    :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

    .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
    """
    def __new__(cls, microseconds: float) -> "Time":
        """
        A class for representing time in microseconds. The class is a subclass of Time, and the value passed in the
        constructor argument will be converted from to microseconds to construct a Time instance.

        :param microseconds: The time in microseconds.
        :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

        .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
        """
        return Time.__new__(cls, int(microseconds))


class Milliseconds(Time):
    """
    A class for representing time in milliseconds. The class is a subclass of Time, and the value passed in the
    constructor argument will be converted from milliseconds to microseconds when constructing an instance of
    :class:`Time <RF_PROBE.timing.Time>`.

    :param milliseconds: The time in milliseconds.
    :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

    .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
    """
    def __new__(cls, milliseconds: float) -> "Time":
        """
        A class for representing time in milliseconds. The class is a subclass of Time, and the value passed in the
        constructor argument will be converted from milliseconds to microseconds when constructing an instance of
        :class:`Time <RF_PROBE.timing.Time>`.

        :param milliseconds: The time in milliseconds.
        :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

        .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
        """
        micros = int(milliseconds * Time._MICROS_PER_MILLISECOND)
        return Time.__new__(cls, micros)


class Seconds(Time):
    """
    A class for representing time in seconds. The class is a subclass of Time, and the value passed in the
    constructor argument will be converted from seconds to microseconds when constructing an instance of
    :class:`Time <RF_PROBE.timing.Time>`.

    :param seconds: The time in seconds
    :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

    .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
    """
    def __new__(cls, seconds: float) -> "Time":
        """
        A class for representing time in seconds. The class is a subclass of Time, and the value passed in the
        constructor argument will be converted from seconds to microseconds when constructing an instance of
        :class:`Time <RF_PROBE.timing.Time>`.

        :param seconds: The time in seconds
        :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

        .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
        """
        micros = int(seconds * Time._MICROS_PER_SECOND)
        return Time.__new__(cls, micros)


class Minutes(Time):
    """
    A class for representing time in minutes. The class is a subclass of Time, and the value passed in the
    constructor argument will be converted from minutes to microseconds when constructing an instance of
    :class:`Time <RF_PROBE.timing.Time>`.

    :param minutes: The time in minutes
    :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

    .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
    """
    def __new__(cls, minutes: float) -> "Time":
        """
        A class for representing time in minutes. The class is a subclass of Time, and the value passed in the
        constructor argument will be converted from minutes to microseconds when constructing an instance of
        :class:`Time <RF_PROBE.timing.Time>`.

        :param minutes: The time in minutes
        :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

        .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
        """
        micros = int(minutes * Time._MICROS_PER_MINUTE)
        return Time.__new__(cls, micros)


class Hours(Time):
    """
    A class for representing time in hours. The class is a subclass of Time, and the value passed in the
    constructor argument will be converted from hours to microseconds when constructing an instance of
    :class:`Time <RF_PROBE.timing.Time>`.

    :param hours: The time in hours
    :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

    .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
    """
    def __new__(cls, hours: float) -> "Time":
        """
        A class for representing time in hours. The class is a subclass of Time, and the value passed in the
        constructor argument will be converted from hours to microseconds when constructing an instance of
        :class:`Time <RF_PROBE.timing.Time>`.

        :param hours: The time in hours
        :returns: An instance of :class:`Time <RF_PROBE.timing.Time>` representing the time in microseconds.

        .. seealso:: :class:`Time <RF_PROBE.timing.Time>`
        """
        micros = int(hours * Time._MICROS_PER_HOUR)
        return Time.__new__(cls, micros)



import subprocess
import sys
import pytest
import os
from pathlib import Path
import toml
from neurobeam.tools.color_scheme import FORMAT_TERMINAL


def retrieve_details(path):
    details = toml.load(path).get("project")
    name = details.get("name")
    version = details.get("version")
    dependencies = details.get("dependencies")
    return name, version, dependencies


def retrieve_project_directory(path):
    return Path(path).parent


def retrieve_project_file():
    project_file = os.path.join(os.getcwd(), "pyproject.toml")
    if not os.path.exists(project_file):
        project_file = os.path.join(Path(os.getcwd()).parent, "pyproject.toml")
    if not os.path.exists(project_file):
        raise FileNotFoundError("Can't find project file")
    return project_file


def collect_project():
    project_file = retrieve_project_file()
    project_directory = retrieve_project_directory(project_file)
    package_name, package_version, package_dependencies = retrieve_details(project_file)
    return project_directory, project_file, package_name, package_version, package_dependencies


# get project information and work from correct directory
proj_dir, proj_file, pkg_name, pkg_version, pkg_dependencies = collect_project()
os.chdir(proj_dir)


print(FORMAT_TERMINAL("\nPackage: ", "emphasis") + FORMAT_TERMINAL(f"{pkg_name}", "type"))
print(FORMAT_TERMINAL(f"Version: ", "emphasis") + f"{pkg_version}")
print(FORMAT_TERMINAL(f"Dependencies: ", "emphasis") + f"{pkg_dependencies}")


@pytest.mark.parametrize("path", [proj_dir])
def test_install(path):
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "install", "-e ."])
    except subprocess.CalledProcessError as e:
        print(f"{e.output}")
        # This doesn't work on non-windows systems, but we don't support those anyway.



from .configs import Config, load_config
from .create import create_controller
from . import extensions

# import to perform registration
from . import adapters  # noqa: F401 // Adapter Registration
from . import behavior  # noqa: F401 // Behavior & Component Registration
from . import callbacks  # noqa: F401 // Callback Registration
from . import hardware  # noqa: F401 // Component Registration
from .tools import audio  # noqa: F401 // Audio Registration


# These are the things we want to explicitly expose to user. By using __all__ we can control what is imported when
# using the * syntax (of course), but also some IDEs will use this to provide autocompletion.
__all__ = [
    "Config",
    "create_controller",
    "extensions",
    "load_config",
]


# Confirm that user is using windows and warn them if they are using legacy windows versions
from .exceptions import LegacyWindowsWarning as _LegacyWindowsWarning
from .exceptions import OperatingSystemUnsupportedError as _OperatingSystemUnsupportedError
from platform import platform as _platform
from warnings import warn as _warn


__platform = _platform()
if "windows" not in __platform.lower():  # pragma: no cover
    raise _OperatingSystemUnsupportedError(__platform)
if not any((version in __platform for version in ["8", "10", "11"])):  # pragma: no cover
    _warn(_LegacyWindowsWarning(__platform))



import pytest

from neurobeam.data import EventLog, RingBuffer, TimestampedRingBuffer, Timestamping
from neurobeam.exceptions import (DataReorganizationWarning, PathMismatchWarning, RingSizeMismatchWarning,
                                  EmptyEventLogWarning, OverloadedEventLogWarning)
import numpy as np


def test_event_log(tmp_path):
    log = EventLog("event_log", path=tmp_path, ring_size=5)
    for _ in range(100):
        log.log_event("TEST", "LOGGING")
    log2 = EventLog(path=tmp_path, ring_size=5)
    log2.log_event("TEST", "LOGGING")
    log.close()
    log2.close()
    del log
    del log2

    # test warnings
    with pytest.warns(PathMismatchWarning):
        log = EventLog("event_log", path=tmp_path, ring_size=5)
        log2 = EventLog("event_log", path=tmp_path.with_name("beans"))
        del log
        del log2
    with pytest.warns(RingSizeMismatchWarning):
        log = EventLog("event_log", path=tmp_path, ring_size=5)
        log2 = EventLog("event_log", path=tmp_path, ring_size=10)
        del log
        del log2
    with pytest.warns(EmptyEventLogWarning):
        log = EventLog("event_log", path=tmp_path)
        log.log_event()
        del log
    with pytest.warns(OverloadedEventLogWarning):
        log = EventLog("event_log", path=tmp_path)
        log.log_event("a", "b", "c", "d")

    # open second unrelated log
    log2 = EventLog("behavior", path=tmp_path, ring_size=10)
    assert len(log) != len(log2), "Event logs are not separate"


@pytest.mark.parametrize("element_depth_", [1, 5, 10])
@pytest.mark.parametrize("element_length_", [1, 5, 10])
@pytest.mark.parametrize("ring_size_", [1, 5, 10])
@pytest.mark.parametrize("data_type_", [np.uint8, np.uint16, np.uint32, np.uint64,
                                        np.int8, np.int16, np.int32, np.int64,
                                        np.float16, np.float32, np.float64])
def test_ring_buffer(tmp_path, helper, element_depth_, element_length_, ring_size_, data_type_):
    # instantiate
    file_path = tmp_path.joinpath("ring_buffer_test")
    rbf = RingBuffer(element_depth=element_depth_,
                     element_length=element_length_,
                     ring_size=ring_size_,
                     path=file_path,
                     dtype=data_type_)
    data = [np.full((element_depth_, element_length_), block, dtype=data_type_) for block in range(25)]

    # test utilization
    for block in data:
        rbf.add_data(block)

    # test get_data calls
    np.testing.assert_array_equal(rbf(), rbf.get_data(), "Mismatched get_data calls")

    # test closing
    rbf.close()

    # test reorganization
    org_data = np.memmap(rbf.org_path, shape=rbf.data_shape, dtype=rbf.data_type)
    np.testing.assert_equal(np.hstack(data), org_data)

    # test print
    with helper.block_printing():
        print(rbf)

    # test clear
    rbf.clear()
    assert (len(rbf) == 0), "Uncleared buffer"

    # test shape exception handling
    with pytest.warns(DataReorganizationWarning):
        rbf._validate_raw(np.ones((42,)))


@pytest.mark.parametrize("element_depth_", [1, 5, 10])
@pytest.mark.parametrize("element_length_", [1, 5, 10])
@pytest.mark.parametrize("ring_size_", [1, 5, 10])
@pytest.mark.parametrize("data_type_", [np.uint8, np.uint16, np.uint32, np.uint64,
                                        np.int8, np.int16, np.int32, np.int64,
                                        np.float16, np.float32, np.float64])
def test_timestamped_ringbuffer(tmp_path, helper, element_depth_, element_length_, ring_size_, data_type_):
    # instantiate
    file_path = tmp_path.joinpath("ring_buffer_test")
    rbf = TimestampedRingBuffer(element_depth=element_depth_,
                                element_length=element_length_,
                                ring_size=ring_size_,
                                path=file_path,
                                dtype=data_type_)
    data = [np.full((element_depth_, element_length_), block, dtype=data_type_) for block in range(25)]

    # test utilization
    for block in data:
        rbf.add_data(block)

    # test get_data calls
    np.testing.assert_array_equal(rbf(), rbf.get_data(), "Mismatched get_data calls")
    assert (len(rbf) == len(rbf.get_timestamps())), "Mismatched timestamps"

    # test closing
    rbf.close()

    # test reorganization
    org_data = np.memmap(rbf.org_path, shape=rbf.data_shape, dtype=rbf.data_type)
    np.testing.assert_equal(np.hstack(data), org_data)

    # test print
    with helper.block_printing():
        print(rbf)

    # test clear
    rbf.clear()
    assert (len(rbf) == 0), "Uncleared buffer"
    assert (len(rbf.timestamps) == 0), "Uncleared timestamps"

    # test shape exception handling
    with pytest.warns(DataReorganizationWarning):
        rbf._validate_raw(np.ones((42,)))


@pytest.mark.parametrize("ring_size_", [1, 5, 10])
def test_timestamping(tmp_path, helper, ring_size_):
    # instantiate
    file_path = tmp_path.joinpath("ring_buffer_test")
    rbf = Timestamping(path=file_path, ring_size=ring_size_)

    # test utilization
    for _ in range(25):
        rbf.add_data()

    # test get_data calls
    np.testing.assert_array_equal(rbf(), rbf.get_data(), "Mismatched get_data calls")

    # test closing
    rbf.close()

    # test print
    with helper.block_printing():
        print(rbf)

    # test clear
    rbf.clear()
    assert (len(rbf) == 0), "Uncleared buffer"

    # test shape exception handling
    with pytest.warns(DataReorganizationWarning):
        rbf._validate_raw(np.ones((42,)))



from inspect import isclass, isfunction, signature
from typing import Callable, Union, Any, Optional, Type, Tuple
from typing_extensions import Protocol, runtime_checkable
import importlib

from .configs import Config
from .timing import Time
from .tools import HAS, EMPTY


"""
Protocols & Custom Types for structural typing of various neurobeam objects during runtime type-checking and for reference 
by users to generate custom neurobeam objects. These protocols are used to define the expected methods and attributes of 
each respective neurobeam object.
"""


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MANUAL TYPE CHECKING
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


def check_protocol(cls: Any, protocol: Union[Type, Tuple[Type, ...]], exception: Optional[Callable] = None) -> bool:
    """
    Check if a class adheres to a particular protocol

    :param cls: The class to check
    :param protocol: The protocol to check against
    :param exception: The exception to raise if the check fails
    :returns: Whether the class adheres to the protocol
    """

    def check_one(cls_: Any, one_protocol: Type) -> bool:
        failed = False
        if isclass(cls):
            for attr in vars(one_protocol):
                if attr not in vars(cls) and not attr.startswith("_"):
                    failed = True
        else:
            failed = True
        if failed and HAS(exception):
            raise exception(cls)
        elif failed and EMPTY(exception):
            return False
        else:
            return True

    if isinstance(protocol, tuple):
        return any((check_one(cls, one_protocol) for one_protocol in protocol))
    else:
        return check_one(cls, protocol)
# TODO: Clean up & Debug


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CALLBACKS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


class _MetaCallback(type):

    @staticmethod
    def _check_is_function(callback: Callable) -> bool:
        """
        Check if callback is a function

        :param function: The callback function to check
        :returns: Whether the callback is a function
        """
        return isfunction(callback)

    @staticmethod
    def _check_parameters(callback: Callable) -> bool:
        """
        Check if the callback has the correct parameters

        :param function: The callback function to check
        :returns: Whether the callback has the correct parameters
        """
        parameters = signature(callback).parameters
        #try:
        #    assert len(parameters) >= 1, "Callback must have at least one parameter"
        #    assert list(parameters.keys())[0] in ("self", "origin"), "First parameter must be self"
        #except AssertionError:
        #    return False
        return True

    def __instancecheck__(self, instance: Any) -> bool:
        """
        Fakes instance check so that functions with the correct parameters can be returned as instances of
        :class:`CallbackFunction <neurobeam.extensions.CallbackFunction>`

        :param instance: The instance to check
        :returns: Whether the instance is an instance of :class:`CallbackFunction <neurobeam.extensions.CallbackFunction>`

        """
        if self._check_is_function(instance) and self._check_parameters(instance):
            return isinstance(instance, object)
        else:
            super().__instancecheck__(instance)


class CallbackFunction(metaclass=_MetaCallback):
    """
    All callback functions must be of the form `def callback(origin, **kwargs) -> Any`. This class is used to define the
    Used for structural typing of callback functions. Here, we define the expected structure of a callback function
    type. Recall, that even functions are objects in python. We could just subclass function, but that would make it
    confusing for users to implement their own callbacks. Instead, we essentially use this class to fake a
    runtime-checkable protocol.
    """

    def __call__(self, *args, **kwargs):
        ...


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// COMPONENTS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@runtime_checkable
class Component(Protocol):
    """
    Protocol for structural typing of neurobeam components. All neurobeam components are expected to have the included
    methods and attributes.
    """

    #: Config: Configuration object containing a sub-configuration for the component (if necessary)
    config: Config

    def build(self) -> None:
        """
        Build the component. This method should be used to initialize the component and set up any necessary resources.

        """
        ...

    def start(self) -> None:
        """
        Start the component. This method should be used to begin any necessary processes or threads that the component
        requires to function and subsequently launch the component's event loop if necessary.

        """
        ...

    def stop(self) -> None:
        """
        Stop the component. This method should be used to halt any necessary processes or threads that the component
        requires to function and subsequently close the component's event loop if necessary.

        """
        ...

    def destroy(self) -> None:
        """
        Destroy the component. This method should be used to clean up any resources that the component has acquired
        during its lifetime. This method is required, but in most cases can be left as 'pass'.

        """
        ...


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// BEHAVIOR
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@runtime_checkable
class BehaviorState(Protocol):
    """
    Protocol for structural typing of neurobeam behavior states. All neurobeam behavior states are expected to have the included
    methods and attributes.
    """
    name: str

    def state_loop(self, time: float) -> bool:
        ...

    def check_progress(self, time: float) -> float:
        ...

    def check_transition(self, time: float) -> str:
        ...

    def initialize(self) -> None:
        ...

    def transition(self) -> None:
        ...


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// ADAPTERS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@runtime_checkable
class Adapter(Protocol):
    """
    Base protocol for structural typing of neurobeam adapters. That is, all neurobeam adapters are expected to have the
    following methods and attributes. Note, this is a base protocol that is not meant to directly guide user
    implementations. Instead, it is meant to be used as a base for other protocols that are more specific to the
    interface in question.
    """
    #: str: name of the interface
    name: str
    #: Config: Configuration object containing a sub-configuration for the component (if necessary)
    config: Config
    #: component request
    component: Any

    def initialized(self):
        ...


class Reader(Adapter):
    """
    Protocol for structural typing of neurobeam read adapters. All neurobeam read adapters are expected to have the
    following methods and attributes.

    """
    #: Union[Callable, CallbackRequest]: Either a request for or an implementation of a callback that will be called to
    #: read a value from the interface.
    def read(self) -> Any:
        ...


class Writer(Adapter):
    """
    Protocol for structural typing of neurobeam write adapters. All neurobeam write adapters are expected to have the
    following methods and attributes.

    """
    #: Union[Callable, CallbackRequest]: Either a request for or an implementation of a callback that will be called
    #: to write a specific value to the interface.
    #command: Union[Callable]

    def command(self, value: Any) -> None:
        ...

    #: Union[Callable, CallbackRequest]: Either a request for or an implementation of a callback that will be called to
    #: set the value written to the interface at the next write.
    def put(self, value: Any) -> None:
        ...

    #: Union[Callable, CallbackRequest]: Either a request for or an implementation of a callback that will be called
    #: to write a pre-specified value to the interface.
    def write(self, value: Any) -> None:
        ...


class Trigger(Adapter):
    """
    Protocol for structural typing of neurobeam trigger adapters. All neurobeam trigger adapters are expected to have the
    following methods and attributes.

    """
    #: Union[Callable, CallbackRequest]: Either a request for or an implementation of a callback that will be called to
    #: trigger a command callback given a specific operation on a read callback
    def command(self, value: Any) -> None:
        ...

    def read(self) -> Any:
        ...

    def trigger(self) -> None:
        ...


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DATA BUFFERS
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@runtime_checkable
class DataBuffer(Protocol):
    """
    Protocol for structural typing of neurobeam data buffers. All neurobeam data buffers are expected to simultaneously write
    the data to file on each addition of data. If you are creating a data buffer that only holds the data in memory,
    then the open and close methods can simply be left as pass.
    """

    # noinspection PyPropertyDefinition
    @property
    def is_open(self) -> bool:
        """
        :Getter: Whether the associated file is open.
        :Getter Type: :class:`bool`
        :Setter: This property is read-only.
        """
        ...

    def add_data(self, data: Any) -> "DataBuffer":
        """
        Add data to the buffer. This method should be used to add data to the buffer. The data can be of any type.

        :param data: The data to be added to the buffer
        """
        ...

    def clear(self) -> "DataBuffer":
        """
        Clear the buffer. This method should be used to clear the buffer of all data.
        """
        ...

    def close(self) -> "DataBuffer":
        """
        Close the associated file for writing data.
        """
        ...

    def get_data(self) -> Any:
        """
        Get data from the buffer. This method should be used to retrieve data from the buffer.

        :returns: The data from the buffer
        """
        ...

    def open(self) -> "DataBuffer":
        """
        Open the associated file for writing data.
        """
        ...


"""
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// REGISTRIES
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
"""


@runtime_checkable
class ClassRegistry(Protocol):
    @staticmethod
    def type_check(obj: Any, raise_exception: bool = True) -> Union[None, bool]:
        ...

    @classmethod
    def register(cls, alias: Optional[str] = None) -> Callable:
        ...

    @classmethod
    def has(cls, name: str) -> bool:
        ...

    @classmethod
    def get(cls, name: str, approximate: bool = False) -> object:
        ...


@runtime_checkable
class InstanceRegistry(Protocol):
    @classmethod
    def register(cls) -> Callable:
        ...

    @classmethod
    def has(cls, name: str, wildcard: Optional[Union[Tuple[str, Any], Tuple[Tuple[str, Any], ...]]] = None) -> bool:
        ...

    @classmethod
    def get(cls, name: str, wildcard: Optional[Union[Tuple[str, Any], Tuple[Tuple[str, Any], ...]]] = None) -> object:
        ...



from abc import ABC, abstractmethod
from typing import Tuple, Optional, Callable
from functools import wraps
import socket
import warnings

# from calscipy.bruker import PrairieLink

from ..configs import HoistedConfig
from ..exceptions import SingletonError
from ..registries import Provider
from ..tools import HAS

# If fast PrairieView implementation is compiled...
try:
    import probe_prairie
except ModuleNotFoundError:
#    warnings.warn("Could not find a compiled probe_prairie library...", UserWarning, stacklevel=2)
#    raise ModuleNotFoundError("Could not find a compiled probe_prairie library..."
#                             "probe_prairie comes pre-compiled for Windows 10+ x64; please raise an issue if you are"
#                              "on a Windows 10+ x64 system. Otherwise, make sure to compile the library!")
    pass


def _format_pv_command(function: Callable) -> Callable:
    @wraps(function)
    def decorator(self, message, *args, **kwargs):
        message += "\r\n"
        message = message.encode()
        return function(self, message, *args, **kwargs)
    return decorator


def _format_pv_response(function: Callable) -> Callable:
    @wraps(function)
    def decorator(self, *args, **kwargs):
        response = function(self, *args, **kwargs).decode()
        response = response.split("\r\n")
        return response[1]
    return decorator


@Provider()
class Microscope:

    __instance = None

    def __new__(cls: "Microscope", *args, **kwargs) -> "Microscope":
        if HAS(cls.__instance):
            raise SingletonError(cls)
        else:
            cls.__instance = super(Microscope, cls).__new__(cls)
        return cls.__instance

    def __init__(self,
                 config: Optional["HoistedConfig"] = None,
                 software: str = "PrairieView",
                 realtime_mode: bool = False,
                 motion_correction: bool = False,
                 identify_rois: bool = False,
                 extract_fluorescence: bool = False,
                 infer_spikes: bool = False,
                 monitor_ensembles: bool = False) -> None:
        """
        Microscope

        :param config: The configuration with a hoisted microscope sub-configuration
        """
        self.config = config
        self.software = config.software if HAS(config) else software
        self.realtime_mode = config.realtime_mode if HAS(config) else realtime_mode
        self.motion_correction = config.motion_correction if HAS(config) else motion_correction
        self.identify_rois = config.identify_rois if HAS(config) else identify_rois
        self.extract_fluorescence = config.extract_fluorescence if HAS(config) else extract_fluorescence
        self.infer_spikes = config.infer_spikes if HAS(config) else infer_spikes
        self.monitor_ensembles = config.monitor_ensembles if HAS(config) else monitor_ensembles

        self._addr = "169.254.239.135"
        self._port = 1236
        self._timeout = 0.01
        self._client = None
        # self._client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # self._client.connect((self._addr, self._port))
        # self._client.settimeout(self._timeout)

    def __str__(self) -> str:
        return f"Microscope: {self.software}, Realtime Mode: {self.realtime_mode}"

    def build(self) -> None:
        pass

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    def destroy(self) -> None:
        pass

    def connected(self) -> bool:
        """
        Whether the microscope is connected to the python interpreter

        """
        ...

    @_format_pv_command
    def command(self, message: str) -> bool:
        self._client.sendall(message)
        response = self._client.recv(16).decode()
        response = response.split("\r\n")
        return any(("ACK" in response_ for response_ in response))

    @_format_pv_response
    def listen(self) -> str:
        return self._client.recv(4096)



from typing import Union, Any
import warnings

from .exceptions import (InvalidAdapterWarning, InvalidComponentWarning, SingletonError, EmptyControllerWarning,
                         EmptyControllerError, ControllerImmutabilityWarning)
from .extensions import Component, Adapter, Reader, Writer, Trigger
from .tools import HAS, EMPTY
from .registries import CallbackRegistry, Provider


@Provider()
class Controller:
    """
    Controller class for neurobeam. The controller is a singleton object that references all the components necessary for
    the neurobeam to function. The controller is responsible for starting, stopping, and destroying the components within
    the controller. The instance of the controller is held in a weak reference as a name-mangled class attribute to
    ensure that only one instance of the controller class is ever active. This singleton-ness was implemented to
    ensure users only ever run one experiment per interpreter and not due to some specific implementation requirement.

    """
    #: Union[Tuple[Component, ...], List[Component]]: List of components within the controller; tuple when controller
    # is locked
    components = []

    #: Union[Tuple[Adapter, ...], List[Adapter]]: List of adapters within the controller; tuple when controller
    # is locked
    adapters = []

    #: bool: Whether the controller is running
    _running = False

    def __init__(self, *args) -> None:
        ################################################################################################################
        # REFERENCE COMPONENTS & adapterS
        ################################################################################################################
        if len(args) == 0:
            warnings.warn(EmptyControllerWarning(), stacklevel=2)
        for arg in args:
            if isinstance(arg, Component):
                self.add_component(arg)
            elif isinstance(arg, (Adapter, Reader, Writer, Trigger)):
                self.add_adapter(arg)
            else:  # If invalid, just raise a warning for both
                warnings.warn(InvalidComponentWarning(arg), stacklevel=2)
                warnings.warn(InvalidAdapterWarning(arg), stacklevel=2)

    def __str__(self) -> str:
        return (f"Controller containing the following components and adapters: "
                f"{[thing.__name__() for thing in self.components + self.adapters]}")

    @property
    def running(self) -> bool:
        """
        :Getter: Returns whether the controller is running.
        :Getter Type: :class:`bool`
        :Setter: This property is read-only
        :Setter Type: :class:`bool`
        """
        return self._running

    @staticmethod
    def __name__() -> str:
        return "Controller"

    @running.setter
    def running(self, value: bool) -> None:
        if value == self._running:
            pass
        elif value:
            # convert to tuple for faster access and immutability
            self.components = tuple(self.components)
            self.adapters = tuple(self.adapters)
            self._running = True
        else:
            # convert to list to allow mutability
            self.components = list(self.components)
            self.adapters = list(self.adapters)
            self._running = False

    @CallbackRegistry.register(alias="controller_add_component")
    def add_component(self, component: Component) -> None:
        """
        Adds a component to the controller.

        :param component: A component to add to the controller
        :raises InvalidComponentWarning: Raised if the component is not a valid component
        :raises ControllerImmutabilityWarning: Raised if the controller is running
        """
        if not isinstance(component, Component):
            warnings.warn(InvalidComponentWarning(component), stacklevel=2)
        elif self.running:
            warnings.warn(ControllerImmutabilityWarning(), stacklevel=2)
        else:
            self.components.append(component)

    @CallbackRegistry.register(alias="controller_add_adapter")
    def add_adapter(self, adapter: Union[Reader, Writer, Trigger]) -> None:
        """
        Adds an adapter to the controller.

        :param adapter: An adapter to add to the controller
        :raises InvalidAdapterWarning: Raised if the adapter is not a valid adapter
        :raises ControllerImmutabilityWarning: Raised if the controller is running
        """
        if not isinstance(adapter, (Adapter, Reader, Writer, Trigger)):
            warnings.warn(InvalidAdapterWarning(adapter), stacklevel=2)
        elif self.running:
            warnings.warn(ControllerImmutabilityWarning(), stacklevel=2)
        else:
            self.adapters.append(adapter)

    @CallbackRegistry.register()
    def start(self) -> None:
        """
        Start the controller. This method should be used to begin any necessary processes or threads that the
        components within the controller require to function.

        """
        if not self.running:
            # noinspection PyCallingNonCallable
            self.running = True
            if EMPTY(self.components + self.adapters):
                raise EmptyControllerError()
            for component in self.components:
                print(f"{component} is starting...\r\r")
                component.start()
                print(f"{component} is starting...success.")

    @CallbackRegistry.register()
    def stop(self) -> None:
        """
        Stop the controller. This method should be used to halt any necessary processes or threads that the
        components within the controller require to function.

        """
        if self.running:
            self.running = False
            if EMPTY(self.components + self.adapters):
                raise EmptyControllerError()
            for component in self.components:
                print(f"{component} is stopping...\r\r")
                component.stop()
                print(f"{component} is stopping...success.")

    @CallbackRegistry.register()
    def destroy(self) -> None:
        """
        Destroy the controller. This method should be used to destroy any necessary processes, threads, or other
        resources that the components within the controller require to function.

        """
        if self.running:
            self.stop()
        if EMPTY(self.components + self.adapters):
            raise EmptyControllerError()
        print("Destroying adapters...\r\r")
        self.adapters = []
        print("Destroying adapters...success.")
        for component in self.components:
            print(f"{component} is destroying...\r\r")
            component.destroy()
            print(f"{component} is destroying...success.")
        self.clear()

    @CallbackRegistry.register()
    def clear(self) -> None:
        if self.running:
            self.destroy()

        self.components = []
        self.adapters = []

    def __del__(self):
        if self.running:
            self.destroy()

    def __enter__(self) -> "Controller":
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()
        self.destroy()



from .conditional_dispatch import conditional_dispatch
from .color_scheme import COLORS, FORMAT_TERMINAL
from .counter import counter
from .devices import DeviceClass, find_devices
from .pattern_matching import PatternMatching
from .readability import HAS, EMPTY, parse_empties, prune_keys
from .shuffling import random_shuffle, balanced_shuffle
from .validators import convert_permitted_types_to_required, validate_filename, validate_extension


# here are things we actually want to expose
__all__ = [
    "balanced_shuffle",
    "COLORS",
    "conditional_dispatch",
    "convert_permitted_types_to_required",
    "counter",
    "DeviceClass",
    "EMPTY",
    "find_devices",
    "FORMAT_TERMINAL",
    "HAS",
    "parse_empties",
    "PatternMatching",
    "prune_keys",
    "random_shuffle",
    "validate_extension",
    "validate_filename",
]



from pathlib import Path
import os


"""
A collection of paths for included files (e.g., default configurations, test paths, font files, etc)
"""


#: Path: The base path of the package
BASE_PATH = Path.cwd()

#: Path: The path to these static files
STATIC_PATH = Path(os.path.realpath(__file__)).parents[0]

#: Path: The path to the neurobeam package
PACKAGE_PATH = STATIC_PATH.parents[0]

#: Path: The path to the neurobeam's default audio intensity look-up table
LUT_PATH = STATIC_PATH.joinpath("LUT.csv")

#: Path: The path to the tests directory
TEST_PATH = PACKAGE_PATH.parents[0].joinpath("tests")

#: Path: The path to the test configurations
TEST_CONFIG_PATH = TEST_PATH.joinpath("test_configs")

#: Path: The path to the gui font file
GUI_FONT_PATH = STATIC_PATH.joinpath("MyriadPro-Regular.tff")

#: Path: The path to neurobeam's large logo
LARGE_ICON_PATH = STATIC_PATH.joinpath("large_logo.ico")

#: Path: The path to neurobeam's small logo
SMALL_ICON_PATH = STATIC_PATH.joinpath("small_logo.ico")

#: Path: The path to the default viewer configuration
DEFAULT_GUI_SETTINGS = STATIC_PATH.joinpath("default_viewer_configuration.ini")

#: Path: The path to the default save path
DEFAULT_SAVE_PATH = BASE_PATH.joinpath("probe_experiments")
if not DEFAULT_SAVE_PATH.exists():  # pragma: no cover
    DEFAULT_SAVE_PATH.mkdir(parents=True)

#: Tuple: Supported Microscopes
SUPPORTED_MICROSCOPES = ("PrairieView", "Thorlabs")
